from concurrent.futures import Executor
from threading import Event
from unittest.mock import ANY, AsyncMock, MagicMock, PropertyMock, call, patch

import pytest
from blueapi.client.event_bus import BlueskyStreamingError
from blueapi.core import BlueskyContext
from blueapi.service.model import TaskRequest
from blueapi.worker import TaskStatus, WorkerState
from blueapi.worker.event import TaskError, TaskResult
from bluesky import RunEngine, RunEngineInterrupted
from bluesky import plan_stubs as bps

from mx_bluesky.common.parameters.components import (
    MxBlueskyParameters,
    get_param_version,
)
from mx_bluesky.common.parameters.constants import Status
from mx_bluesky.common.utils.exceptions import CrystalNotFoundError, SampleError
from mx_bluesky.hyperion._plan_runner_params import (
    RobotUnload,
    UDCCleanup,
    UDCDefaultState,
    Wait,
)
from mx_bluesky.hyperion.blueapi.parameters import LoadCentreCollectParams
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect
from mx_bluesky.hyperion.plan_runner import PlanError
from mx_bluesky.hyperion.supervisor import SupervisorRunner

TEST_VISIT = "cm12345-67"


@pytest.fixture
def mock_bluesky_context(run_engine: RunEngine):
    return BlueskyContext(run_engine=run_engine)


@pytest.fixture(autouse=True)
def mock_blueapi_client():
    with patch(
        "mx_bluesky.hyperion.supervisor._supervisor.BlueapiClient"
    ) as mock_class:
        mock_client = mock_class.from_config.return_value
        yield mock_client


@pytest.fixture(autouse=True)
def mock_blueapi_client_state(mock_blueapi_client):
    mock_state = PropertyMock(return_value=WorkerState.IDLE)
    type(mock_blueapi_client).state = mock_state
    yield mock_state


@pytest.fixture
def blueapi_config():
    return MagicMock()


@pytest.fixture
def runner(mock_bluesky_context, blueapi_config):
    runner = SupervisorRunner(mock_bluesky_context, blueapi_config, True)
    with patch.object(runner, "check_external_callbacks_are_alive"):
        yield runner


@patch("mx_bluesky.hyperion.supervisor._supervisor.TaskMonitor")
def test_decode_and_execute_load_centre_collect_executes_and_monitors_the_task_and_returns_the_visit(
    mock_task_monitor: MagicMock,
    mock_blueapi_client: MagicMock,
    runner: SupervisorRunner,
    external_load_centre_collect_params,
):
    parent = MagicMock()
    parent.attach_mock(mock_blueapi_client, "blueapi_client")
    parent.attach_mock(mock_task_monitor, "TaskMonitor")
    parent.attach_mock(mock_task_monitor.return_value, "task_monitor")

    result = runner.context.run_engine(
        runner.decode_and_execute(TEST_VISIT, [external_load_centre_collect_params])
    )

    expected_task_request = TaskRequest(
        name="load_centre_collect",
        params={"parameters": external_load_centre_collect_params},
        instrument_session=TEST_VISIT,
    )
    parent.assert_has_calls(
        [
            call.TaskMonitor(mock_blueapi_client, expected_task_request),
            call.task_monitor.__enter__(),
            call.blueapi_client.run_task(
                expected_task_request,
                on_event=mock_task_monitor.return_value.__enter__.return_value.on_blueapi_event,
            ),
            call.task_monitor.__exit__(None, None, None),
        ]
    )

    assert result.plan_result == TEST_VISIT  # type: ignore


def test_decode_and_execute_wait(
    mock_blueapi_client: MagicMock, runner: SupervisorRunner
):
    mock_sleep = AsyncMock()
    runner.context.run_engine.register_command("sleep", mock_sleep)
    try:
        runner.context.run_engine(
            runner.decode_and_execute(
                TEST_VISIT,
                [
                    Wait.model_validate(
                        {
                            "parameter_model_version": get_param_version(),
                            "duration_s": 10,
                        }
                    )
                ],
            )
        )
    finally:
        runner.context.run_engine.register_command(
            "sleep", runner.context.run_engine._sleep
        )

    mock_blueapi_client.run_task.assert_not_called()
    mock_sleep.assert_awaited_once()


def test_decode_and_execute_default_state(
    mock_blueapi_client: MagicMock, runner: SupervisorRunner
):
    runner.context.run_engine(
        runner.decode_and_execute(
            TEST_VISIT,
            [
                UDCDefaultState.model_validate(
                    {"parameter_model_version": get_param_version()}
                )
            ],
        )
    )

    mock_blueapi_client.run_task.assert_called_once_with(
        TaskRequest(
            name="move_to_udc_default_state", params={}, instrument_session=TEST_VISIT
        ),
        on_event=ANY,
    )


def test_decode_and_execute_robot_unload(
    mock_blueapi_client: MagicMock, runner: SupervisorRunner
):
    runner.context.run_engine(runner.decode_and_execute(TEST_VISIT, [RobotUnload()]))

    mock_blueapi_client.run_task.assert_called_once_with(
        TaskRequest(
            name="robot_unload",
            params={"visit": TEST_VISIT},
            instrument_session=TEST_VISIT,
        ),
        on_event=ANY,
    )


def test_decode_and_execute_udc_cleanup(
    mock_blueapi_client: MagicMock, runner: SupervisorRunner
):
    runner.context.run_engine(
        runner.decode_and_execute(
            TEST_VISIT,
            [
                UDCCleanup.model_validate(
                    {"parameter_model_version": get_param_version()}
                )
            ],
        )
    )

    mock_blueapi_client.run_task.assert_called_once_with(
        TaskRequest(
            name="clean_up_udc",
            params={},
            instrument_session=TEST_VISIT,
        ),
        on_event=ANY,
    )


def test_current_status_set_to_busy_during_execution(
    mock_blueapi_client: MagicMock, runner: SupervisorRunner, executor: Executor
):
    task_executing = Event()
    check_complete = Event()

    def check_busy():
        task_executing.wait(timeout=1)
        assert runner.current_status == Status.BUSY
        check_complete.set()

    def wait_for_check(*args, **kwargs):
        task_executing.set()
        check_complete.wait(1)
        return TaskStatus(
            task_id="TASK_ID",
            result=TaskResult(result=None, type="None"),
            task_complete=True,
            task_failed=False,
        )

    mock_blueapi_client.run_task.side_effect = wait_for_check

    fut = executor.submit(check_busy)
    try:
        assert runner.current_status == Status.IDLE
        runner.context.run_engine(
            runner.decode_and_execute(
                TEST_VISIT,
                [
                    UDCDefaultState.model_validate(
                        {"parameter_model_version": get_param_version()}
                    )
                ],
            )
        )
    finally:
        fut.result(1)


def test_current_status_set_to_failed_on_exception_and_raise_plan_error(
    mock_blueapi_client: MagicMock, runner: SupervisorRunner
):
    status = TaskStatus(
        task_id="TASK_ID",
        result=TaskError(type="RuntimeError", message="Simulated exception"),
        task_complete=True,
        task_failed=True,
    )
    mock_blueapi_client.run_task.return_value = status

    with pytest.raises(
        PlanError, match="Exception raised.* message='Simulated exception'"
    ):
        runner.context.run_engine(
            runner.decode_and_execute(
                TEST_VISIT,
                [
                    UDCDefaultState.model_validate(
                        {"parameter_model_version": get_param_version()}
                    )
                ],
            )
        )
    assert runner.current_status == Status.FAILED


def test_is_connected_queries_blueapi_client(
    runner: SupervisorRunner,
    mock_blueapi_client: MagicMock,
    mock_blueapi_client_state: PropertyMock,
):
    assert runner.is_connected()
    mock_blueapi_client_state.assert_called_once()


def test_is_connected_returns_false_on_exception(
    runner: SupervisorRunner,
    mock_blueapi_client: MagicMock,
    mock_blueapi_client_state: PropertyMock,
):
    mock_blueapi_client_state.side_effect = RuntimeError("Simulated exception")
    assert not runner.is_connected()


def test_shutdown_sends_abort_to_run_engine_when_idle(
    runner: SupervisorRunner, executor: Executor
):
    runner.run_engine = MagicMock(spec=RunEngine)

    def request_shutdown():
        runner.shutdown()

    fut = executor.submit(request_shutdown)

    fut.result(1)
    runner.run_engine.abort.assert_called_once()


def test_shutdown_sends_abort_to_blueapi_client_when_running_then_aborts(
    runner: SupervisorRunner, executor: Executor
):
    task_running = Event()
    remote_abort_requested = Event()

    def request_shutdown():
        task_running.wait(1)
        runner.shutdown()

    def mock_run_task(*args, **kwargs):
        task_running.set()
        remote_abort_requested.wait(1)
        raise BlueskyStreamingError("Simulated abort exception")

    def mock_baton_handler_loop():
        yield from runner.decode_and_execute(
            TEST_VISIT,
            [
                UDCDefaultState.model_validate(
                    {"parameter_model_version": get_param_version()}
                )
            ],
        )
        while True:
            yield from bps.sleep(0.1)

    runner.blueapi_client.run_task.side_effect = mock_run_task  # type: ignore
    runner.blueapi_client.abort.side_effect = remote_abort_requested.set  # type: ignore

    fut = executor.submit(request_shutdown)
    try:
        with pytest.raises(RunEngineInterrupted):
            runner.context.run_engine(mock_baton_handler_loop())
    finally:
        fut.result(1)


def test_exception_during_callback_check_raises_plan_error(
    runner: SupervisorRunner, load_centre_collect_params: LoadCentreCollect
):
    with patch.object(
        runner,
        "check_external_callbacks_are_alive",
        side_effect=RuntimeError("Simulated exception"),
    ):
        with pytest.raises(PlanError, match="Exception raised.*: Simulated exception"):
            runner.context.run_engine(
                runner.decode_and_execute(TEST_VISIT, [load_centre_collect_params])
            )


def test_unrecognised_instruction_raises_assertion_error(runner: SupervisorRunner):
    with pytest.raises(AssertionError, match="Unsupported instruction"):
        runner.context.run_engine(
            runner.decode_and_execute(
                TEST_VISIT,
                [
                    MxBlueskyParameters.model_validate(
                        {"parameter_model_version": get_param_version()}
                    )
                ],
            )
        )


@pytest.mark.parametrize(
    "exception_type", [SampleError.__name__, CrystalNotFoundError.__name__]
)
def test_sample_error_skips_subsequent_instructions(
    runner: SupervisorRunner,
    external_load_centre_collect_params: LoadCentreCollectParams,
    mock_blueapi_client: MagicMock,
    exception_type: str,
):
    mock_blueapi_client.run_task.return_value = TaskStatus(
        task_id="TASK_ID",
        result=TaskError(type=exception_type, message="Simulated sample error"),
        task_complete=True,
        task_failed=True,
    )
    runner.context.run_engine(
        runner.decode_and_execute(
            TEST_VISIT,
            [external_load_centre_collect_params, external_load_centre_collect_params],
        )
    )
    mock_blueapi_client.run_task.assert_called_once_with(ANY, on_event=ANY)



from unittest.mock import patch

from bluesky.callbacks import CallbackBase
from bluesky_stomp.models import MessageTopic
from event_model import Event, RunStart, RunStop

from mx_bluesky.hyperion.external_interaction.callbacks.__main__ import (
    StompDispatcherContextMgr,
    main,
)
from mx_bluesky.hyperion.parameters.cli import CallbackArgs

CALLBACK_TOPIC = "callback_test_events"


class PatchedStompCallbackMgr(StompDispatcherContextMgr):
    def __init__(self, args: CallbackArgs, callbacks: list[CallbackBase]) -> None:
        super().__init__(args, callbacks)
        callback = callbacks[0]
        assert isinstance(callback, StompTestCallback)
        self.callback = callback

    def __enter__(self):
        super().__enter__()
        self.callback.init(self._stomp_client)
        return self


class StompTestCallback(CallbackBase):
    def __init__(self) -> None:
        super().__init__()
        self.stomp_client = None
        self.destination = MessageTopic(name=CALLBACK_TOPIC)

    def init(self, stomp_client):
        self.stomp_client = stomp_client
        self.fire_event_back("init")

    def start(self, doc: RunStart) -> RunStart | None:
        self.fire_event_back(f"start: {doc['run_name']}")  # type: ignore
        return super().start(doc)

    def stop(self, doc: RunStop) -> RunStop | None:
        self.fire_event_back("stop")
        return super().stop(doc)

    def event(self, doc: Event) -> Event:
        self.fire_event_back(f"event: {doc['data']['baton-requested_user']}")
        return super().event(doc)

    def fire_event_back(self, msg: str):
        self.stomp_client.send(destination=self.destination, obj=msg)  # type: ignore


if __name__ == "__main__":

    def mock_setup_callbacks():
        return [StompTestCallback()]

    with (
        patch(
            "mx_bluesky.hyperion.external_interaction.callbacks.__main__.setup_callbacks",
            return_value=mock_setup_callbacks(),
        ),
        patch(
            "mx_bluesky.hyperion.external_interaction.callbacks.__main__.StompDispatcherContextMgr",
            PatchedStompCallbackMgr,
        ),
    ):
        main()



from functools import partial
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from bluesky.utils import Msg
from dodal.devices.backlight import InOut
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.thawer import OnOff
from dodal.devices.webcam import Webcam
from ophyd_async.core import completed_status, set_mock_value

from mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy import (
    RobotLoadAndEnergyChangeComposite,
    robot_load_and_change_energy_plan,
    take_robot_snapshots,
)
from mx_bluesky.hyperion.external_interaction.callbacks.robot_actions.ispyb_callback import (
    RobotLoadISPyBCallback,
)
from mx_bluesky.hyperion.parameters.robot_load import RobotLoadAndEnergyChange

from ....conftest import raw_params_from_file


@pytest.fixture
def robot_load_and_energy_change_params(tmp_path):
    params = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_robot_load_params.json",
        tmp_path,
    )
    return RobotLoadAndEnergyChange(**params)


@pytest.fixture
def robot_load_and_energy_change_params_no_energy(robot_load_and_energy_change_params):
    robot_load_and_energy_change_params.demand_energy_ev = None
    return robot_load_and_energy_change_params


def dummy_set_energy_plan(energy, composite):
    return (yield Msg("set_energy_plan"))


@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy.set_energy_plan",
    MagicMock(side_effect=dummy_set_energy_plan),
)
def test_when_plan_run_with_requested_energy_specified_energy_change_executes(
    robot_load_and_energy_change_composite: RobotLoadAndEnergyChangeComposite,
    robot_load_and_energy_change_params: RobotLoadAndEnergyChange,
    sim_run_engine: RunEngineSimulator,
):
    sim_run_engine.add_handler(
        "read",
        lambda msg: {"dcm-energy_in_keV": {"value": 11.105}},
        "dcm-energy_in_keV",
    )
    messages = sim_run_engine.simulate_plan(
        robot_load_and_change_energy_plan(
            robot_load_and_energy_change_composite, robot_load_and_energy_change_params
        )
    )
    assert_message_and_return_remaining(
        messages, lambda msg: msg.command == "set_energy_plan"
    )


def run_simulating_smargon_wait(
    robot_load_then_centre_params,
    robot_load_composite,
    total_disabled_reads,
    sim_run_engine: RunEngineSimulator,
):
    num_of_reads = 0

    def return_not_disabled_after_reads(_):
        nonlocal num_of_reads
        num_of_reads += 1
        return {"values": {"value": int(num_of_reads < total_disabled_reads)}}

    sim_run_engine.add_handler(
        "locate",
        lambda msg: {"readback": 11.105},
        "dcm-energy_in_keV",
    )
    sim_run_engine.add_handler(
        "read", return_not_disabled_after_reads, "smargon-disabled"
    )

    with patch(
        "mx_bluesky.common.device_setup_plans.robot_load_unload.SLEEP_PER_CHECK", 1
    ):
        return sim_run_engine.simulate_plan(
            robot_load_and_change_energy_plan(
                robot_load_composite, robot_load_then_centre_params
            )
        )


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.robot_actions.ispyb_callback.ExpeyeInteraction"
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy.set_energy_plan",
    MagicMock(return_value=iter([])),
)
def test_given_ispyb_callback_attached_when_robot_load_then_centre_plan_called_then_ispyb_deposited(
    exp_eye: MagicMock,
    robot_load_and_energy_change_composite: RobotLoadAndEnergyChangeComposite,
    robot_load_and_energy_change_params: RobotLoadAndEnergyChange,
    run_engine: RunEngine,
):
    robot = robot_load_and_energy_change_composite.robot
    webcam = robot_load_and_energy_change_composite.webcam
    set_mock_value(
        robot_load_and_energy_change_composite.oav.snapshot.last_saved_path,
        "test_oav_snapshot",
    )
    set_mock_value(webcam.last_saved_path, "test_webcam_snapshot")
    webcam.trigger = MagicMock(side_effect=lambda: completed_status())
    set_mock_value(robot.barcode, "BARCODE")

    run_engine.subscribe(RobotLoadISPyBCallback())

    action_id = 1098
    exp_eye.return_value.start_robot_action.return_value = action_id

    run_engine(
        robot_load_and_change_energy_plan(
            robot_load_and_energy_change_composite, robot_load_and_energy_change_params
        )
    )

    exp_eye.return_value.start_robot_action.assert_called_once_with(
        "LOAD", "cm31105", 4, 12345
    )
    exp_eye.return_value.update_robot_action.assert_called_once_with(
        action_id,
        {
            "sampleBarcode": "BARCODE",
            "xtalSnapshotBefore": "test_webcam_snapshot",
            "xtalSnapshotAfter": "test_oav_snapshot",
            "containerLocation": 3,
            "dewarLocation": 40,
        },
    )
    exp_eye.return_value.end_robot_action.assert_called_once_with(
        action_id, "success", "OK"
    )


@patch("mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy.datetime")
async def test_when_take_snapshots_called_then_filename_and_directory_set_and_device_triggered(
    mock_datetime: MagicMock, oav: OAV, webcam: Webcam, run_engine: RunEngine
):
    test_directory = "TEST"

    mock_datetime.now.return_value.strftime.return_value = "TIME"

    oav.snapshot.trigger = MagicMock(side_effect=oav.snapshot.trigger)
    webcam.trigger = MagicMock(side_effect=lambda: completed_status())

    run_engine(take_robot_snapshots(oav, webcam, Path(test_directory)))

    oav.snapshot.trigger.assert_called_once()
    assert await oav.snapshot.filename.get_value() == "TIME_oav-snapshot_after_load"
    assert await oav.snapshot.directory.get_value() == test_directory

    webcam.trigger.assert_called_once()
    assert (await webcam.filename.get_value()) == "TIME_webcam_after_load"
    assert (await webcam.directory.get_value()) == test_directory


def test_given_lower_gonio_moved_when_robot_load_then_lower_gonio_moved_to_home_and_back(
    robot_load_and_energy_change_composite: RobotLoadAndEnergyChangeComposite,
    robot_load_and_energy_change_params_no_energy: RobotLoadAndEnergyChange,
    sim_run_engine: RunEngineSimulator,
):
    initial_values = {"z": 0.13, "x": 0.11, "y": 0.12}

    def get_read(axis, msg):
        return {"readback": initial_values[axis]}

    for axis in initial_values.keys():
        sim_run_engine.add_handler(
            "locate", partial(get_read, axis), f"lower_gonio-{axis}"
        )

    messages = sim_run_engine.simulate_plan(
        robot_load_and_change_energy_plan(
            robot_load_and_energy_change_composite,
            robot_load_and_energy_change_params_no_energy,
        )
    )

    for axis in initial_values.keys():
        messages = assert_message_and_return_remaining(
            messages,
            lambda msg: msg.command == "set"
            and msg.obj.name == f"lower_gonio-{axis}"
            and msg.args == (0,),
        )

    for axis, initial in initial_values.items():
        messages = assert_message_and_return_remaining(
            messages,
            lambda msg: msg.command == "set"
            and msg.obj.name == f"lower_gonio-{axis}"
            and msg.args == (initial,),
        )


@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy.set_energy_plan",
    MagicMock(return_value=iter([])),
)
def test_when_plan_run_then_lower_gonio_moved_before_robot_loads_and_back_after_smargon_enabled(
    robot_load_and_energy_change_composite: RobotLoadAndEnergyChangeComposite,
    robot_load_and_energy_change_params_no_energy: RobotLoadAndEnergyChange,
    sim_run_engine: RunEngineSimulator,
):
    initial_values = {"z": 0.13, "x": 0.11, "y": 0.12}

    def get_read(axis, msg):
        return {"readback": initial_values[axis]}

    for axis in initial_values.keys():
        sim_run_engine.add_handler(
            "locate", partial(get_read, axis), f"lower_gonio-{axis}"
        )

    messages = sim_run_engine.simulate_plan(
        robot_load_and_change_energy_plan(
            robot_load_and_energy_change_composite,
            robot_load_and_energy_change_params_no_energy,
        )
    )

    assert_message_and_return_remaining(
        messages, lambda msg: msg.command == "set" and msg.obj.name == "robot"
    )

    for axis in initial_values.keys():
        messages = assert_message_and_return_remaining(
            messages,
            lambda msg: msg.command == "set"
            and msg.obj.name == f"lower_gonio-{axis}"
            and msg.args == (0,),
        )

    for axis, initial in initial_values.items():
        messages = assert_message_and_return_remaining(
            messages,
            lambda msg: msg.command == "set"
            and msg.obj.name == f"lower_gonio-{axis}"  # noqa
            and msg.args == (initial,),  # noqa
        )


@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy.set_energy_plan",
    MagicMock(return_value=iter([])),
)
def test_when_plan_run_then_thawing_turned_on(
    robot_load_and_energy_change_composite: RobotLoadAndEnergyChangeComposite,
    robot_load_and_energy_change_params_no_energy: RobotLoadAndEnergyChange,
    sim_run_engine: RunEngineSimulator,
):
    sim_run_engine.add_handler(
        "read",
        lambda msg: {"dcm-energy_in_keV": {"value": 11.105}},
        "dcm-energy_in_keV",
    )

    messages = sim_run_engine.simulate_plan(
        robot_load_and_change_energy_plan(
            robot_load_and_energy_change_composite,
            robot_load_and_energy_change_params_no_energy,
        )
    )

    assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "set"
        and msg.obj.name == "thawer"
        and msg.args[0] == OnOff.ON,
    )


@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy.set_energy_plan",
    MagicMock(return_value=iter([])),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy.take_robot_snapshots",
    MagicMock(return_value=iter([Msg("take_robot_snapshots")])),
)
def test_when_plan_run_then_backlight_moved_in_before_snapshots_taken(
    robot_load_and_energy_change_composite: RobotLoadAndEnergyChangeComposite,
    robot_load_and_energy_change_params_no_energy: RobotLoadAndEnergyChange,
    sim_run_engine: RunEngineSimulator,
):
    sim_run_engine.add_handler(
        "read",
        lambda msg: {"dcm-energy_in_keV": {"value": 11.105}},
        "dcm-energy_in_keV",
    )

    messages = sim_run_engine.simulate_plan(
        robot_load_and_change_energy_plan(
            robot_load_and_energy_change_composite,
            robot_load_and_energy_change_params_no_energy,
        )
    )

    msgs = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "set"
        and msg.obj.name == "backlight"
        and msg.args[0] == InOut.IN,
    )

    backlight_move_group = msgs[0].kwargs.get("group")

    msgs = assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "wait"
        and msg.kwargs["group"] == backlight_move_group,
    )

    assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "take_robot_snapshots"
    )



import asyncio
from unittest.mock import patch

import pydantic
from bluesky import plan_stubs as bps
from bluesky.preprocessors import run_decorator, set_run_key_decorator
from bluesky.utils import MsgGenerator
from dodal.common import inject
from dodal.devices.aperturescatterguard import ApertureScatterguard
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.motors import XYZStage
from dodal.devices.robot import BartRobot
from dodal.devices.smargon import Smargon
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.log import LOGGER as DODAL_LOGGER
from dodal.log import set_up_stream_handler
from ophyd_async.core import (
    observe_value,
    set_mock_value,
)

from mx_bluesky.common.preprocessors.preprocessors import (
    pause_xbpm_feedback_during_collection_at_desired_transmission_decorator,
)
from mx_bluesky.common.utils.exceptions import CrystalNotFoundError, WarningError
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion.blueapi.parameters import (
    HyperionParam,
    LoadCentreCollectParams,
)
from mx_bluesky.hyperion.experiment_plans.load_centre_collect_full_plan import (
    LoadCentreCollectComposite,
)

BEAMLINE_ERROR_SAMPLE_ID = 111111
CRYSTAL_NOT_FOUND_SAMPLE_ID = 222222


def publish_event(plan_name: str):
    yield from bps.open_run(md={"plan_name": plan_name})
    yield from bps.close_run()


def load_centre_collect(
    parameters: LoadCentreCollectParams,
    composite: LoadCentreCollectComposite = inject(),
) -> MsgGenerator:
    LOGGER.info(f"load_centre_collect called with sample id {parameters.sample_id}")
    if parameters.sample_id == BEAMLINE_ERROR_SAMPLE_ID:
        raise ValueError("Simulated beamline error")
    elif parameters.sample_id == CRYSTAL_NOT_FOUND_SAMPLE_ID:
        raise CrystalNotFoundError("Simulated crystal not found")
    yield from bps.sleep(1)


def robot_unload(
    visit: str,
    cleanup_group: str = "cleanup",
    robot: BartRobot = inject("robot"),
    smargon: Smargon = inject("gonio"),
    aperture_scatterguard: ApertureScatterguard = inject("aperture_scatterguard"),
    lower_gonio: XYZStage = inject("lower_gonio"),
) -> MsgGenerator:
    yield from publish_event("robot_unload")
    match visit:
        case "raise_warning_error":
            raise WarningError("Test warning error")
        case "raise_other_error":
            raise RuntimeError("Test unexpected error")
        case "wait_for_abort":
            while True:
                yield from bps.sleep(1)


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class WaitForFeedbackComposite:
    xbpm_feedback: XBPMFeedback
    attenuator: BinaryFilterAttenuator


class WaitForFeedbackParams(HyperionParam):
    time_for_beam_stable: float
    time_in_plan: float
    visit: str
    sample_id: int
    sample_puck: int


def wait_for_feedback(
    parameters: WaitForFeedbackParams,
    devices: WaitForFeedbackComposite = inject(),
) -> MsgGenerator:
    LOGGER.info("wait_for_feedback plan called...")
    set_mock_value(devices.xbpm_feedback.baton_ref().commissioning, False)  # type: ignore
    set_mock_value(devices.xbpm_feedback.pos_stable, 0)

    async def become_stable():
        await asyncio.sleep(parameters.time_for_beam_stable)
        set_mock_value(devices.xbpm_feedback.pos_stable, 1)

    real_observe = observe_value

    async def patched_observe(signal):
        wait_for_stable = asyncio.create_task(become_stable())
        async for _ in real_observe(signal):
            yield _
        await wait_for_stable

    @pause_xbpm_feedback_during_collection_at_desired_transmission_decorator(
        devices=devices, desired_transmission_fraction=1.0
    )
    @set_run_key_decorator("wait_for_feedback")
    @run_decorator()
    def inner_plan() -> MsgGenerator:
        LOGGER.info("Inner plan called...")
        yield from bps.sleep(parameters.time_in_plan)
        LOGGER.info("Finished waiting")

    with patch(
        "dodal.devices.xbpm_feedback.observe_value", side_effect=patched_observe
    ):
        yield from inner_plan()


set_up_stream_handler(DODAL_LOGGER)



from pathlib import Path
from unittest.mock import ANY, AsyncMock, MagicMock, patch

import bluesky.plan_stubs as bps
import pytest
from bluesky.run_engine import RunEngine
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from ophyd_async.core import set_mock_value

from mx_bluesky.beamlines.aithre_lasershaping.experiment_plans.robot_load_plan import (
    RobotLoadComposite,
    _take_robot_snapshots,
    robot_load_and_snapshots_plan,
    robot_unload_plan,
)
from mx_bluesky.beamlines.aithre_lasershaping.parameters.constants import CONST
from mx_bluesky.beamlines.aithre_lasershaping.parameters.robot_load_parameters import (
    AithreRobotLoad,
)
from mx_bluesky.beamlines.aithre_lasershaping.robot_load import (
    robot_load_and_snapshot,
    robot_unload,
)
from mx_bluesky.hyperion.external_interaction.callbacks.robot_actions.ispyb_callback import (
    RobotLoadISPyBCallback,
)

from ....conftest import raw_params_from_file


@pytest.fixture
def robot_load_params(tmp_path):
    params = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_robot_load_params.json",
        tmp_path,
    )
    return AithreRobotLoad(**params)


@pytest.fixture
def aithre_robot_load_composite(
    robot,
    aithre_gonio,
    oav,
) -> RobotLoadComposite:
    composite = RobotLoadComposite(
        robot,
        aithre_gonio,
        oav,
        aithre_gonio,
    )
    return composite


@pytest.fixture
def mock_pin_tip(pin_tip: PinTipDetection):
    pin_tip._get_tip_and_edge_data = AsyncMock(retrun_value=pin_tip.INVALID_POSITION)
    return pin_tip


def noop_plan():
    yield from bps.null()


@patch(
    "mx_bluesky.beamlines.aithre_lasershaping.experiment_plans.robot_load_plan.pin_tip_centre_plan",
    side_effect=lambda *args, **kwargs: noop_plan(),
)
@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.robot_actions.ispyb_callback.ExpeyeInteraction"
)
def test_given_ispyb_callback_attached_when_robot_load_and_snapshots_plan_called_then_ispyb_deposited(
    exp_eye: MagicMock,
    mock_pin_tip_centring_plan,
    mock_pin_tip: PinTipDetection,
    aithre_robot_load_composite: RobotLoadComposite,
    robot_load_params: AithreRobotLoad,
    run_engine: RunEngine,
):
    set_mock_value(
        aithre_robot_load_composite.oav.snapshot.last_saved_path,
        "test_oav_snapshot",
    )
    set_mock_value(aithre_robot_load_composite.robot.barcode, "BARCODE")

    run_engine.subscribe(RobotLoadISPyBCallback())

    action_id = 1098
    exp_eye.return_value.start_robot_action.return_value = action_id

    run_engine(
        robot_load_and_snapshots_plan(
            aithre_robot_load_composite, robot_load_params, mock_pin_tip
        )
    )

    mock_pin_tip_centring_plan.assert_called_once()
    exp_eye.return_value.start_robot_action.assert_called_once_with(
        "LOAD", "cm31105", 4, 12345
    )
    exp_eye.return_value.update_robot_action.assert_called_once_with(
        action_id,
        {
            "sampleBarcode": "BARCODE",
            "xtalSnapshotAfter": "test_oav_snapshot",
            "containerLocation": 3,
            "dewarLocation": 40,
        },
    )
    exp_eye.return_value.end_robot_action.assert_called_once_with(
        action_id, "success", "OK"
    )


@patch(
    "mx_bluesky.beamlines.aithre_lasershaping.experiment_plans.robot_load_plan.datetime"
)
async def test_when_take_snapshots_called_then_filename_and_directory_set_and_device_triggered(
    mock_datetime: MagicMock, oav: OAV, run_engine: RunEngine
):
    test_directory = "TEST"

    mock_datetime.now.return_value.strftime.return_value = "TIME"

    oav.snapshot.trigger = MagicMock(side_effect=oav.snapshot.trigger)

    run_engine(_take_robot_snapshots(oav, Path(test_directory)))

    oav.snapshot.trigger.assert_called_once()
    assert await oav.snapshot.filename.get_value() == "TIME_oav-snapshot_after_load"
    assert await oav.snapshot.directory.get_value() == test_directory


@patch(
    "mx_bluesky.beamlines.aithre_lasershaping.experiment_plans.robot_load_plan._move_gonio_to_home_position",
    autospec=True,
)
async def test_when_robot_unload_called_then_sample_area_prepared_before_load(
    mock_move_gonio,
    aithre_robot_load_composite: RobotLoadComposite,
    robot_load_params: AithreRobotLoad,
    run_engine: RunEngine,
):
    run_engine(robot_unload_plan(aithre_robot_load_composite, robot_load_params))

    mock_move_gonio.assert_called_once()


def test_when_unload_plan_run_then_initial_unload_ispyb_deposition_made(
    aithre_robot_load_composite: RobotLoadComposite,
    robot_load_params: AithreRobotLoad,
    run_engine: RunEngine,
):
    callback = RobotLoadISPyBCallback()
    callback.expeye = (mock_expeye := MagicMock())
    run_engine.subscribe(callback)

    run_engine(robot_unload_plan(aithre_robot_load_composite, robot_load_params))

    mock_expeye.start_robot_action.assert_called_once_with(
        "UNLOAD", "cm31105", 4, 12345
    )


def test_when_unload_plan_run_then_full_ispyb_deposition_made(
    aithre_robot_load_composite: RobotLoadComposite,
    robot_load_params: AithreRobotLoad,
    run_engine: RunEngine,
):
    callback = RobotLoadISPyBCallback()
    callback.expeye = (mock_expeye := MagicMock())
    run_engine.subscribe(callback)

    set_mock_value(
        aithre_robot_load_composite.oav.snapshot.last_saved_path,
        "test_oav_snapshot",
    )
    set_mock_value(aithre_robot_load_composite.robot.current_pin, 3)
    set_mock_value(aithre_robot_load_composite.robot.current_puck, 40)

    action_id = 1098
    mock_expeye.start_robot_action.return_value = action_id

    run_engine(robot_unload_plan(aithre_robot_load_composite, robot_load_params))

    mock_expeye.start_robot_action.assert_called_once_with(
        "UNLOAD", "cm31105", 4, 12345
    )
    mock_expeye.update_robot_action.assert_called_once_with(
        action_id,
        {
            "sampleBarcode": "BARCODE",
            "xtalSnapshotAfter": "test_oav_snapshot",
            "containerLocation": 3,
            "dewarLocation": 40,
        },
    )
    mock_expeye.end_robot_action.assert_called_once_with(action_id, "success", "OK")


@patch(
    "mx_bluesky.beamlines.aithre_lasershaping.experiment_plans.robot_load_plan._move_gonio_to_home_position",
    autospec=True,
)
def test_when_unload_plan_fails_then_error_deposited_in_ispyb(
    mock_move_gonio: MagicMock,
    aithre_robot_load_composite: RobotLoadComposite,
    robot_load_params: AithreRobotLoad,
    run_engine: RunEngine,
):
    class TestError(Exception): ...

    callback = RobotLoadISPyBCallback()
    callback.expeye = (mock_expeye := MagicMock())
    run_engine.subscribe(callback)
    mock_move_gonio.side_effect = TestError("Bad Error")

    action_id = 1098
    mock_expeye.start_robot_action.return_value = action_id

    with pytest.raises(TestError):
        run_engine(robot_unload_plan(aithre_robot_load_composite, robot_load_params))

    mock_expeye.start_robot_action.assert_called_once_with("UNLOAD", "cm31105", 4, ANY)
    mock_expeye.end_robot_action.assert_called_once_with(action_id, "fail", "Bad Error")


@patch(
    "mx_bluesky.beamlines.aithre_lasershaping.experiment_plans.robot_load_plan._robot_load_and_snapshots",
    autospec=True,
)
def test_when_robot_load_and_snapshot_plan_called_correct_plan_called(
    mock_robot_load_and_snapshots,
    aithre_robot_load_composite: RobotLoadComposite,
    robot_load_params: AithreRobotLoad,
    run_engine: RunEngine,
):
    tip_offset = 0
    oav_config = CONST.OAV_CENTRING_FILE
    mock_pin_tip_detection = MagicMock(spec=PinTipDetection)
    run_engine(
        robot_load_and_snapshot(
            aithre_robot_load_composite.robot,
            aithre_robot_load_composite.gonio,
            aithre_robot_load_composite.oav,
            mock_pin_tip_detection,
            tip_offset,
            oav_config,
            robot_load_params.sample_puck or 0,
            robot_load_params.sample_pin or 0,
            robot_load_params.sample_id,
            robot_load_params.visit,
        )
    )

    mock_robot_load_and_snapshots.assert_called_once()


@patch(
    "mx_bluesky.beamlines.aithre_lasershaping.experiment_plans.robot_load_plan._move_gonio_to_home_position",
    autospec=True,
)
def test_when_robot_unload_plan_called_correct_plan_called(
    mock_move_gonio,
    aithre_robot_load_composite: RobotLoadComposite,
    robot_load_params: AithreRobotLoad,
    run_engine: RunEngine,
):
    run_engine(
        robot_unload(
            aithre_robot_load_composite.robot,
            aithre_robot_load_composite.gonio,
            aithre_robot_load_composite.oav,
            robot_load_params.sample_puck or 0,
            robot_load_params.sample_pin or 0,
            robot_load_params.sample_id,
            robot_load_params.visit,
        )
    )

    mock_move_gonio.assert_called_once()



from unittest.mock import MagicMock, Mock, call, patch

import pytest
from dodal.devices.zocalo import ZocaloStartInfo

from mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback import (
    ZocaloCallback,
    ZocaloTrigger,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
    StoreInIspyb,
)
from mx_bluesky.common.utils.exceptions import ISPyBDepositionNotMadeError
from mx_bluesky.hyperion.external_interaction.callbacks.__main__ import (
    create_gridscan_callbacks,
)
from mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback import (
    generate_start_info_from_ordered_runs,
)

EXPECTED_RUN_START_MESSAGE = {"subplan_name": "test_plan_name", "uid": "my_uuid"}
EXPECTED_RUN_END_MESSAGE = {"event": "end", "run_start": "my_uuid"}


class TestZocaloHandler:
    def _setup_handler(self):
        zocalo_handler = ZocaloCallback(
            "test_plan_name", "test_env", generate_start_info_from_ordered_runs
        )
        return zocalo_handler

    def test_handler_doesnt_trigger_on_wrong_plan(self):
        zocalo_handler = self._setup_handler()
        zocalo_handler.start({"sybplan_name": "_not_test_plan_name"})  # type: ignore
        assert len(zocalo_handler.zocalo_info) == 0

    @patch(
        "mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback.ZocaloTrigger",
        autospec=True,
    )
    def test_handler_stores_collection_if_ispyb_ids_come_in_with_triggering_plan(
        self, zocalo_trigger: ZocaloTrigger
    ):
        zocalo_handler = self._setup_handler()
        assert not zocalo_handler.zocalo_info
        zocalo_handler.start(
            {
                **EXPECTED_RUN_START_MESSAGE,
                "ispyb_dcids": (135, 139),
                "scan_points": [{"test": [1, 2, 3]}, {"test": [2, 3, 4]}],
            }  # type: ignore
        )
        assert len(zocalo_handler.zocalo_info) == 2

    @patch(
        "mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback.ZocaloTrigger",
        autospec=True,
    )
    def test_handler_stores_collection_ispyb_ids_come_in_as_subplan(
        self, zocalo_trigger: ZocaloTrigger
    ):
        zocalo_handler = self._setup_handler()
        assert not zocalo_handler.zocalo_info
        zocalo_handler.start(EXPECTED_RUN_START_MESSAGE)  # type: ignore
        assert not zocalo_handler.zocalo_info
        zocalo_handler.start(
            {
                "subplan_name": "other_plan",
                "ispyb_dcids": (135,),
                "scan_points": [{"test": [1, 2, 3]}],
            }  # type: ignore
        )
        zocalo_handler.start(
            {
                "subplan_name": "other_plan",
                "ispyb_dcids": (563,),
                "scan_points": [{"test": [2, 3, 4]}],
            }  # type: ignore
        )

        assert len(zocalo_handler.zocalo_info) == 2

    @patch(
        "mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback.ZocaloTrigger",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.external_interaction.callbacks.xray_centre.nexus_callback.NexusWriter",
    )
    @patch(
        "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.StoreInIspyb",
    )
    def test_execution_of_do_fgs_triggers_zocalo_calls(
        self,
        ispyb_store: MagicMock,
        nexus_writer: MagicMock,
        zocalo_trigger,
        test_event_data,
    ):
        dc_ids = (1, 2)
        dcg_id = 4

        mock_ids = IspybIds(data_collection_ids=dc_ids, data_collection_group_id=dcg_id)
        ispyb_store.return_value.mock_add_spec(StoreInIspyb)

        _, ispyb_cb = create_gridscan_callbacks()
        ispyb_cb.active = True
        assert isinstance(zocalo_handler := ispyb_cb.emit_cb, ZocaloCallback)
        zocalo_handler._reset_state()
        zocalo_handler._reset_state = MagicMock()

        ispyb_store.return_value.begin_deposition.return_value = mock_ids
        ispyb_store.return_value.update_deposition.return_value = mock_ids

        ispyb_cb.start(test_event_data.test_grid_detect_and_gridscan_start_document)  # type: ignore
        ispyb_cb.descriptor(test_event_data.test_descriptor_document_oav_snapshot)
        ispyb_cb.event(test_event_data.test_event_document_oav_snapshot_xy)
        ispyb_cb.event(test_event_data.test_event_document_oav_snapshot_xz)
        ispyb_cb.start(test_event_data.test_gridscan_outer_start_document)  # type: ignore
        ispyb_cb.start(test_event_data.test_do_fgs_start_document)  # type: ignore
        ispyb_cb.descriptor(
            test_event_data.test_descriptor_document_pre_data_collection
        )  # type: ignore
        ispyb_cb.event(test_event_data.test_event_document_pre_data_collection)
        ispyb_cb.descriptor(test_event_data.test_descriptor_document_zocalo_hardware)
        ispyb_cb.event(test_event_data.test_event_document_zocalo_hardware)
        ispyb_cb.descriptor(
            test_event_data.test_descriptor_document_during_data_collection  # type: ignore
        )
        ispyb_cb.event(test_event_data.test_event_document_during_data_collection)
        assert zocalo_handler.zocalo_interactor is not None

        expected_start_calls = [
            call(ZocaloStartInfo(1, "test_path", 0, 200, 0)),
            call(ZocaloStartInfo(2, "test_path", 200, 300, 1)),
        ]

        zocalo_handler.zocalo_interactor.run_start.assert_has_calls(  # type: ignore
            expected_start_calls
        )
        assert zocalo_handler.zocalo_interactor.run_start.call_count == len(dc_ids)  # type: ignore

        ispyb_cb.stop(test_event_data.test_do_fgs_stop_document)
        ispyb_cb.stop(test_event_data.test_gridscan_outer_stop_document)

        zocalo_handler.zocalo_interactor.run_end.assert_has_calls(  # type: ignore
            [call(x) for x in dc_ids]
        )
        assert zocalo_handler.zocalo_interactor.run_end.call_count == len(dc_ids)  # type: ignore

        zocalo_handler._reset_state.assert_called()

    @patch(
        "mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback.ZocaloTrigger",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.external_interaction.callbacks.xray_centre.nexus_callback.NexusWriter",
    )
    @patch(
        "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.StoreInIspyb",
    )
    def test_do_fgs_triggers_zocalo_calls_when_snapshots_in_reverse_order(
        self,
        ispyb_store: MagicMock,
        nexus_writer: MagicMock,
        zocalo_trigger,
        test_event_data,
    ):
        dc_ids = (1, 2)
        dcg_id = 4

        mock_ids = IspybIds(data_collection_ids=dc_ids, data_collection_group_id=dcg_id)
        ispyb_store.return_value.mock_add_spec(StoreInIspyb)

        _, ispyb_cb = create_gridscan_callbacks()
        ispyb_cb.active = True
        assert isinstance(zocalo_handler := ispyb_cb.emit_cb, ZocaloCallback)
        zocalo_handler._reset_state()
        zocalo_handler._reset_state = MagicMock()

        ispyb_store.return_value.begin_deposition.return_value = mock_ids
        ispyb_store.return_value.update_deposition.return_value = mock_ids

        ispyb_cb.start(test_event_data.test_grid_detect_and_gridscan_start_document)  # type: ignore
        ispyb_cb.start(test_event_data.test_gridscan_outer_start_document)  # type: ignore
        ispyb_cb.descriptor(test_event_data.test_descriptor_document_oav_snapshot)
        ispyb_cb.event(test_event_data.test_event_document_oav_snapshot_xz)
        ispyb_cb.event(test_event_data.test_event_document_oav_snapshot_xy)
        ispyb_cb.start(test_event_data.test_do_fgs_start_document)  # type: ignore
        ispyb_cb.descriptor(
            test_event_data.test_descriptor_document_pre_data_collection
        )  # type: ignore
        ispyb_cb.event(test_event_data.test_event_document_pre_data_collection)
        ispyb_cb.descriptor(test_event_data.test_descriptor_document_zocalo_hardware)
        ispyb_cb.event(test_event_data.test_event_document_zocalo_hardware)
        ispyb_cb.descriptor(
            test_event_data.test_descriptor_document_during_data_collection  # type: ignore
        )
        ispyb_cb.event(test_event_data.test_event_document_during_data_collection)
        assert zocalo_handler.zocalo_interactor is not None

        expected_start_calls = [
            call(ZocaloStartInfo(1, "test_path", 0, 200, 0)),
            call(ZocaloStartInfo(2, "test_path", 200, 300, 1)),
        ]

        zocalo_handler.zocalo_interactor.run_start.assert_has_calls(  # type: ignore
            expected_start_calls
        )
        assert zocalo_handler.zocalo_interactor.run_start.call_count == len(dc_ids)  # type: ignore
        ispyb_cb.stop(test_event_data.test_do_fgs_stop_document)
        ispyb_cb.stop(test_event_data.test_gridscan_outer_stop_document)

        zocalo_handler.zocalo_interactor.run_end.assert_has_calls(  # type: ignore
            [call(dc_ids[0]), call(dc_ids[1])]
        )
        assert zocalo_handler.zocalo_interactor.run_end.call_count == len(dc_ids)  # type: ignore

    @patch(
        "mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback.ZocaloTrigger",
        autospec=True,
    )
    def test_handler_raises_on_the_end_of_a_plan_with_no_depositions_and_still_resets_generator(
        self, zocalo_trigger: ZocaloTrigger
    ):
        def info_generator():
            _ = yield
            info1 = ZocaloStartInfo(
                ispyb_dcid=1,
                filename="blah",
                start_frame_index=0,
                number_of_frames=10,
                message_index=1,
            )
            yield [info1]

        info_generator_factory = Mock()
        info_generator_factory.side_effect = info_generator

        zocalo_handler = ZocaloCallback(
            "test_plan_name", "test_env", info_generator_factory
        )

        zocalo_handler.start(EXPECTED_RUN_START_MESSAGE)  # type: ignore
        assert info_generator_factory.call_count == 1

        with pytest.raises(ISPyBDepositionNotMadeError):
            zocalo_handler.stop(EXPECTED_RUN_END_MESSAGE)  # type: ignore

        assert info_generator_factory.call_count == 2



import pytest
from dodal.beamlines import i24
from dodal.devices.attenuator.attenuator import EnumFilterAttenuator
from dodal.devices.beamlines.i24.beam_center import DetectorBeamCenter
from ophyd_async.core import set_mock_value

from mx_bluesky.beamlines.i24.serial.fixed_target.ft_utils import ChipType
from mx_bluesky.beamlines.i24.serial.parameters.experiment_parameters import (
    ExtruderParameters,
    FixedTargetParameters,
)
from mx_bluesky.beamlines.i24.serial.parameters.utils import get_chip_format


@pytest.fixture
def eiger_beam_center() -> DetectorBeamCenter:
    bc: DetectorBeamCenter = i24.eiger_beam_center.build(
        connect_immediately=True, mock=True
    )
    set_mock_value(bc.beam_x, 1605)
    set_mock_value(bc.beam_y, 1702)
    return bc


@pytest.fixture
def enum_attenuator() -> EnumFilterAttenuator:
    return i24.attenuator.build(connect_immediately=True, mock=True)


@pytest.fixture
def dummy_params_without_pp():
    oxford_defaults = get_chip_format(ChipType.Oxford)
    params = {
        "visit": "/tmp/dls/i24/fixed/foo",
        "directory": "bar",
        "filename": "chip",
        "exposure_time_s": 0.01,
        "detector_distance_mm": 100,
        "detector_name": "eiger",
        "transmission": 1.0,
        "num_exposures": 1,
        "chip": oxford_defaults.model_dump(),
        "map_type": 1,
        "pump_repeat": 0,
        "checker_pattern": False,
        "chip_map": [1],
    }
    return FixedTargetParameters(**params)  # type: ignore


@pytest.fixture
def dummy_params_ex():
    params = {
        "visit": "/tmp/dls/i24/extruder/foo",
        "directory": "bar",
        "filename": "protein",
        "exposure_time_s": 0.1,
        "detector_distance_mm": 100,
        "detector_name": "eiger",
        "transmission": 1.0,
        "num_images": 10,
        "pump_status": False,
    }
    return ExtruderParameters(**params)  # type: ignore



import dataclasses

import pytest
from dodal.devices.zebra.zebra import (
    I24Axes,
    RotationDirection,
    Zebra,
)
from dodal.devices.zebra.zebra_controlled_shutter import (
    ZebraShutter,
    ZebraShutterControl,
)

from mx_bluesky.common.device_setup_plans.setup_zebra_and_shutter import (
    configure_zebra_and_shutter_for_auto_shutter,
    setup_zebra_for_gridscan,
    setup_zebra_for_rotation,
    tidy_up_zebra_after_gridscan,
)


async def _get_shutter_input_2(zebra: Zebra):
    return (
        await zebra.logic_gates.and_gates[zebra.mapping.AND_GATE_FOR_AUTO_SHUTTER]
        .sources[2]
        .get_value()
    )


async def _get_shutter_input_1(zebra: Zebra):
    return (
        await zebra.logic_gates.and_gates[zebra.mapping.AND_GATE_FOR_AUTO_SHUTTER]
        .sources[1]
        .get_value()
    )


async def test_configure_zebra_and_shutter_for_auto(
    run_engine, zebra: Zebra, zebra_shutter: ZebraShutter
):
    run_engine(
        configure_zebra_and_shutter_for_auto_shutter(
            zebra, zebra_shutter, zebra.mapping.sources.IN4_TTL
        )
    )
    assert await zebra_shutter.control_mode.get_value() == ZebraShutterControl.AUTO
    assert await _get_shutter_input_1(zebra) == zebra.mapping.sources.SOFT_IN1
    assert await _get_shutter_input_2(zebra) == zebra.mapping.sources.IN4_TTL


async def test_zebra_cleanup(run_engine, zebra: Zebra, zebra_shutter: ZebraShutter):
    run_engine(tidy_up_zebra_after_gridscan(zebra, zebra_shutter, wait=True))
    assert (
        await zebra.output.out_pvs[zebra.mapping.outputs.TTL_DETECTOR].get_value()
        == zebra.mapping.sources.PC_PULSE
    )
    assert await _get_shutter_input_2(zebra) == zebra.mapping.sources.PC_GATE


async def test_zebra_set_up_for_gridscan(
    run_engine, zebra: Zebra, zebra_shutter: ZebraShutter
):
    @dataclasses.dataclass
    class Composite:
        zebra: Zebra
        sample_shutter: ZebraShutter

    composite = Composite(zebra, zebra_shutter)
    run_engine(setup_zebra_for_gridscan(composite, wait=True))
    assert (
        await zebra.output.out_pvs[zebra.mapping.outputs.TTL_DETECTOR].get_value()
        == zebra.mapping.sources.IN3_TTL
    )
    assert await _get_shutter_input_2(zebra) == zebra.mapping.sources.IN4_TTL
    assert await zebra_shutter.control_mode.get_value() == ZebraShutterControl.AUTO
    assert await _get_shutter_input_1(zebra) == zebra.mapping.sources.SOFT_IN1


async def test_zebra_set_up_for_rotation(
    run_engine,
    zebra: Zebra,
    zebra_shutter: ZebraShutter,
):
    axis = I24Axes.OMEGA
    start_angle = 90
    scan_width = 180
    shutter_opening_deg = 1
    shutter_opening_s: float = 0.08
    direction = RotationDirection.NEGATIVE
    ttl_input_for_detector_to_use = 3
    run_engine(
        setup_zebra_for_rotation(
            zebra,
            zebra_shutter,
            axis,
            start_angle,
            scan_width,
            shutter_opening_deg,
            shutter_opening_s,
            direction,
            "group",
            True,
            ttl_input_for_detector_to_use,
        )
    )
    assert await zebra.pc.gate_trigger.get_value() == axis
    assert await zebra.pc.gate_width.get_value() == pytest.approx(
        scan_width + shutter_opening_deg, 0.01
    )
    assert await zebra_shutter.control_mode.get_value() == ZebraShutterControl.AUTO
    assert await _get_shutter_input_1(zebra) == zebra.mapping.sources.SOFT_IN1
    assert await zebra.pc.dir.get_value() == direction
    assert await zebra.pc.gate_start.get_value() == start_angle
    assert await zebra.pc.pulse_start.get_value() == shutter_opening_s
    assert (
        await zebra.output.out_pvs[ttl_input_for_detector_to_use].get_value()
        == zebra.mapping.sources.PC_PULSE
    )



from collections.abc import Callable

import bluesky.plan_stubs as bps
from bluesky import preprocessors as bpp
from bluesky.preprocessors import plan_mutator
from bluesky.utils import Msg, MsgGenerator, make_decorator

from mx_bluesky.common.device_setup_plans.xbpm_feedback import (
    check_and_pause_feedback,
    unpause_xbpm_feedback_and_set_transmission_to_1,
)
from mx_bluesky.common.parameters.constants import PlanNameConstants
from mx_bluesky.common.protocols.protocols import (
    XBPMPauseDevices,
)


def _create_insert_plans_mutator(
    run_key_to_wrap: PlanNameConstants | None = None,
    on_open: Callable[[Msg], MsgGenerator] | None = None,
    on_close: Callable[[], MsgGenerator] | None = None,
) -> Callable[[Msg], tuple[MsgGenerator | None, MsgGenerator | None]]:
    """
    Inserts plans to be executed when the run with the given name opens/closes.

    Args:
        run_key_to_wrap: The run name to insert before/after. If None (default) then
            insert on the first open/close run regardless of its name.
        on_open: The plan to perform just before the run opens
        on_close: The plan to perform just after the run closes

    """
    _wrapped_run_name: None | str = None

    def insert_plans(msg: Msg):
        # Wrap the specified run, or, if none specified, wrap the first run encountered
        nonlocal _wrapped_run_name

        match msg.command:
            case "open_run":
                # If we specified a run key, did we encounter it
                # If we didn't specify, then insert the plans and track the name of the run
                if on_open is not None and (
                    not (run_key_to_wrap or _wrapped_run_name)
                    or run_key_to_wrap is msg.run
                ):
                    _wrapped_run_name = msg.run if msg.run else "unnamed_run"
                    return on_open(msg), None
            case "close_run":
                # Check if the run tracked from above was closed
                # An exception is raised in the RunEngine if two unnamed runs are opened
                # at the same time, so we are safe from unpausing on the wrong run
                if on_close is not None and (
                    (_wrapped_run_name == "unnamed_run" and not msg.run)
                    or (msg.run and _wrapped_run_name and _wrapped_run_name is msg.run)
                ):
                    return None, on_close()

        return None, None

    return insert_plans


def pause_xbpm_feedback_during_collection_at_desired_transmission_wrapper(
    plan: MsgGenerator,
    devices: XBPMPauseDevices,
    desired_transmission_fraction: float,
    run_key_to_wrap: PlanNameConstants | None = None,
):
    """
    Sets the transmission for the data collection, ensuring the xbpm feedback is valid, then pauses
     XBPM feedback. Resets transmission and unpauses XBPM immediately after the run has finished.

    This wrapper should be attached to the entry point of any beamline-specific plan that may disrupt the XBPM feedback,
    such as a data collection or an x-ray center grid scan.
    This wrapper will do nothing if no runs are seen.

    XBPM feedback isn't reliable during collections due to:
     * Objects (e.g. attenuator) crossing the beam can cause large (incorrect) feedback movements
     * Lower transmissions/higher energies are less reliable for the xbpm

    So we need to keep the transmission at 100% and the feedback on when not collecting
    and then turn it off and set the correct transmission for collection. The feedback
    mostly accounts for slow thermal drift so it is safe to assume that the beam is
    stable during a collection.

    Args:
        plan: The plan performing the data collection.
        devices (XBPMPauseDevices): Composite device including The XBPM device that is responsible for keeping
            the beam in position, and attenuator
        desired_transmission_fraction (float): The desired transmission for the collection
        run_key_to_wrap: (str | None): Pausing XBPM and setting transmission is inserted before the 'open_run' message is seen with
            the matching run key, and unpausing and resetting transmission is inserted after the corresponding 'close_run' message is
            seen. If not specified, instead wrap the first run encountered.
    """

    def head(msg: Msg):
        yield from check_and_pause_feedback(
            devices.xbpm_feedback,
            devices.attenuator,
            desired_transmission_fraction,
        )

        # Allow 'open_run' message to pass through
        yield msg

    def tail():
        yield from unpause_xbpm_feedback_and_set_transmission_to_1(
            devices.xbpm_feedback, devices.attenuator
        )

    mutator = _create_insert_plans_mutator(run_key_to_wrap, head, tail)

    # Contingency wrapper can cause unpausing to occur on exception and again on close_run.
    # Not needed after https://github.com/bluesky/bluesky/issues/1891
    return (
        yield from bpp.contingency_wrapper(
            plan_mutator(plan, mutator),
            except_plan=lambda _: unpause_xbpm_feedback_and_set_transmission_to_1(
                devices.xbpm_feedback,
                devices.attenuator,
            ),
        )
    )


def set_transmission_and_trigger_xbpm_feedback_before_collection_wrapper(
    plan: MsgGenerator,
    devices: XBPMPauseDevices,
    desired_transmission_fraction: float,
    run_key_to_wrap: PlanNameConstants | None = None,
):
    """
    Sets the transmission and triggers xbpm feedback immediately before
    doing the specified run. Doesn't revert transmission at the end of the plan.

    Args:
        plan: The plan performing the data collection.
        devices (XBPMPauseDevices): Composite device including The XBPM device that is responsible for keeping
            the beam in position, and the attenuator
        desired_transmission_fraction (float): The desired transmission for the collection
        run_key_to_wrap: (str | None): 'Set transmission' and 'trigger XBPM device' is inserted before the 'open_run' message with
            the matching run key is seen. If not specified, instead wrap the first run encountered.
    """

    def head(msg: Msg):
        yield from bps.mv(devices.attenuator, desired_transmission_fraction)
        yield from bps.trigger(devices.xbpm_feedback)

        # Allow 'open_run' message to pass through
        yield msg

    mutator = _create_insert_plans_mutator(run_key_to_wrap, head)

    return (yield from plan_mutator(plan, mutator))


set_transmission_and_trigger_xbpm_feedback_before_collection_decorator = make_decorator(
    set_transmission_and_trigger_xbpm_feedback_before_collection_wrapper
)


pause_xbpm_feedback_during_collection_at_desired_transmission_decorator = (
    make_decorator(
        pause_xbpm_feedback_during_collection_at_desired_transmission_wrapper
    )
)



"""
Cleaner abstractions of the PV table.

Takes the PV tables from I24's setup_beamline and wraps a slightly more
abstract wrapper around them.
"""

from mx_bluesky.beamlines.i24.serial.setup_beamline import pv


class Eiger:
    id = 94
    name = "eiger"

    pixel_size_mm = (0.075, 0.075)
    image_size_pixels = (3108, 3262)

    image_size_mm = tuple(
        round(a * b, 3) for a, b in zip(image_size_pixels, pixel_size_mm, strict=False)
    )

    det_y_threshold = 70.0
    det_y_target = 59.0

    class PV:
        detector_distance = pv.eiger_detdist
        wavelength = pv.eiger_wavelength
        transmission = "BL24I-EA-PILAT-01:cam1:FilterTransm"
        filename_rbv = pv.eiger_od_filename_rbv
        file_name = pv.eiger_od_filename
        file_path = pv.eiger_od_filepath
        file_template = None
        sequence_id = pv.eiger_seq_id
        beamx = pv.eiger_beamx
        beamy = pv.eiger_beamy
        bit_depth = pv.eiger_bitdepthrbv

    def __str__(self) -> str:
        return self.name


Detector = Eiger



"""This module contains the experimental plans which hyperion can run.

The __all__ list in here are the plans that are externally available from outside Hyperion.
"""

from mx_bluesky.hyperion.experiment_plans.load_centre_collect_full_plan import (
    load_centre_collect_full,
)

__all__ = [
    "load_centre_collect_full",
]



import math
from math import asin

from scanspec.core import AxesPoints, Axis
from scipy.constants import physical_constants

from mx_bluesky.common.utils.log import LOGGER

hc_in_ev_and_angstrom: float = (
    physical_constants["speed of light in vacuum"][0]
    * physical_constants["Planck constant in eV/Hz"][0]
    * 1e10  # Angstroms per metre
)


def interconvert_ev_angstrom(wavelength_or_energy: float) -> float:
    return hc_in_ev_and_angstrom / wavelength_or_energy


def convert_ev_to_angstrom(hv: float) -> float:
    return interconvert_ev_angstrom(hv)


def convert_angstrom_to_ev(wavelength: float) -> float:
    return interconvert_ev_angstrom(wavelength)


def number_of_frames_from_scan_spec(scan_points: AxesPoints[Axis]):
    ax = list(scan_points.keys())[0]
    return len(scan_points[ax])


def energy_to_bragg_angle(energy_kev: float, d_a: float) -> float:
    """Compute the bragg angle given the energy in kev.

    Args:
        energy_kev:  The energy in keV
        d_a:         The lattice spacing in Angstroms
    Returns:
        The bragg angle in degrees
    """
    wavelength_a = convert_ev_to_angstrom(energy_kev * 1000)
    d = d_a
    return asin(wavelength_a / (2 * d)) * 180 / math.pi


def fix_transmission_and_exposure_time_for_current_wavelength(
    current_wavelength_a: float,
    assumed_wavelength_a_from_settings: float,
    requested_trans_frac: float = 100,
    requested_exposure_time_s: float = 0.004,
) -> tuple[float, float]:
    """
    Calculates an exposure time and transmission fraction for XRC which will provide a good signal
    on the detector by using a known good wavelength, comparing it to the beamlines current wavelength,
    then scaling accordingly.

    Args:
        current_wavelength_a: Current energy of the beamline in angstroms.
        assumed_wavelength_a_from_settings: The known "good" wavelength. This should be read from
            'gda.px.expttable.default.wavelength' in GDA's domain.properties, via the config server.
        requested_trans_frac: Requested transmission fraction to use.
        requested_exposure_time_s: Requested exposure time to use.

    Returns:
        The scaled transmission fraction and exposure time respectively, in a tuple.
    """

    wavelength_scale = (assumed_wavelength_a_from_settings / current_wavelength_a) ** 2

    # Transmission frac needed to get ideal signal
    ideal_trans_frac = requested_trans_frac * wavelength_scale
    if ideal_trans_frac <= 1:
        new_trans_frac = ideal_trans_frac
        new_exposure_time_s = requested_exposure_time_s
    else:
        # If the scaling would result in transmission fraction > 1,
        # cap it to 1, find remaining scaling needed, and apply it
        # to exposure time instead.
        new_trans_frac = 1
        scaling_applied_to_trans = new_trans_frac / requested_trans_frac
        remaining_scaling_needed = wavelength_scale / scaling_applied_to_trans
        new_exposure_time_s = requested_exposure_time_s * remaining_scaling_needed

    LOGGER.info(
        f"Fixing transmission fraction to {new_trans_frac} and exposure time to {new_exposure_time_s}s"
    )

    # Exposure time in FGS IOC is in ms, and must be an integer, so round it here
    return new_trans_frac, round(new_exposure_time_s, 3)



from contextlib import nullcontext
from typing import Any
from unittest.mock import MagicMock, patch

import pytest
from bluesky import Msg, RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from dodal.devices.aperturescatterguard import ApertureValue
from dodal.devices.hutch_shutter import ShutterState
from dodal.devices.ipin import IPinGain
from dodal.devices.mx_phase1.beamstop import BeamstopPositions
from dodal.devices.scintillator import InOut
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutterState
from ophyd_async.core import set_mock_value, set_mock_values

from mx_bluesky.common.experiment_plans.beamstop_check import (
    _FEEDBACK_TIMEOUT_S,
    _GROUP_POST_BEAMSTOP_OUT_CHECK,
    _GROUP_PRE_BEAMSTOP_OUT_CHECK,
    BeamObstructedError,
    BeamstopCheckDevices,
    BeamstopNotInPositionError,
    SampleCurrentBelowThresholdError,
    _post_beamstop_out_check_actions,
    move_beamstop_in_and_verify_using_diode,
)


def test_beamstop_check_closes_sample_shutter(
    beamstop_check_devices, sim_run_engine, beamline_parameters
):
    msgs = sim_run_engine.simulate_plan(
        move_beamstop_in_and_verify_using_diode(
            beamstop_check_devices,
            beamline_parameters,
            250,
            800,
        )
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is beamstop_check_devices.sample_shutter
        and msg.args[0] == ZebraShutterState.CLOSE,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait"
        and msg.kwargs["group"] == msgs[0].kwargs["group"],
    )


@patch(
    "mx_bluesky.common.experiment_plans.beamstop_check.unpause_xbpm_feedback_and_set_transmission_to_1"
)
def test_beamstop_check_raises_error_if_feedback_fails_to_stabilise(
    mock_unpause_feedback, beamstop_check_devices, run_engine, beamline_parameters
):
    mock_unpause_feedback.side_effect = TimeoutError
    with pytest.raises(SampleCurrentBelowThresholdError):
        run_engine(
            move_beamstop_in_and_verify_using_diode(
                beamstop_check_devices,
                beamline_parameters,
                250,
                800,
            )
        )

    mock_unpause_feedback.assert_called_once_with(
        beamstop_check_devices.xbpm_feedback,
        beamstop_check_devices.attenuator,
        _FEEDBACK_TIMEOUT_S,
    )


@patch(
    "mx_bluesky.common.experiment_plans.beamstop_check.unpause_xbpm_feedback_and_set_transmission_to_1"
)
def test_beamstop_check_performs_pre_beamstop_out_check_actions_before_first_background_read(
    mock_unpause_feedback,
    beamstop_check_devices: BeamstopCheckDevices,
    sim_run_engine,
    beamline_parameters,
):
    """Check that:
    * sample shutter closed
    * feedback unpaused, xmission 100%
    * ap-sg move to out
    * ipin gain set
    * detector shutter closed
    * beamstop out
    * All the above are complete
    Also check that
    * detector move to in range started"""

    mock_unpause_feedback.return_value = iter([Msg("unpause_feedback")])
    all_msgs = sim_run_engine.simulate_plan(
        move_beamstop_in_and_verify_using_diode(
            beamstop_check_devices,
            beamline_parameters,
            250,
            800,
        )
    )

    # Feedback + Transmission
    msgs = assert_message_and_return_remaining(
        all_msgs, lambda msg: msg.command == "unpause_feedback"
    )

    # Pre-beamstop out checks
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is beamstop_check_devices.backlight
        and msg.args[0] == InOut.OUT,
    )
    pre_check_group = msgs[0].kwargs["group"]
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is beamstop_check_devices.aperture_scatterguard.selected_aperture
        and msg.args[0] == ApertureValue.OUT_OF_BEAM
        and msg.kwargs["group"] == pre_check_group,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is beamstop_check_devices.ipin.gain
        and msg.args[0] == IPinGain.GAIN_10E4_LOW_NOISE
        and msg.kwargs["group"] == pre_check_group,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is beamstop_check_devices.detector_motion.shutter
        and msg.args[0] == ShutterState.CLOSED
        and msg.kwargs["group"] == pre_check_group,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is beamstop_check_devices.beamstop.selected_pos
        and msg.args[0] == BeamstopPositions.OUT_OF_BEAM
        and msg.kwargs["group"] == pre_check_group,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait" and msg.kwargs["group"] == pre_check_group,
    )
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "sleep" and msg.args[0] == 1
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "read"
        and msg.obj is beamstop_check_devices.ipin.pin_readback,
    )


def test_beamstop_check_completes_post_beamstop_out_check_actions_before_second_check(
    beamstop_check_devices, sim_run_engine, beamline_parameters
):
    msgs = sim_run_engine.simulate_plan(
        move_beamstop_in_and_verify_using_diode(
            beamstop_check_devices,
            beamline_parameters,
            250,
            800,
        )
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is beamstop_check_devices.detector_motion.z
        and msg.args[0] == 250,
    )
    post_check_group = msgs[0].kwargs["group"]
    # check background read happens first
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "read"
        and msg.obj is beamstop_check_devices.ipin.pin_readback,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait" and msg.kwargs["group"] == post_check_group,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is beamstop_check_devices.sample_shutter
        and msg.args[0] == ZebraShutterState.OPEN,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait"
        and msg.kwargs["group"] == msgs[0].kwargs["group"],
    )
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "sleep" and msg.args[0] == 1
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "read"
        and msg.obj is beamstop_check_devices.ipin.pin_readback,
    )


def test_beamstop_check_ensures_detector_shutter_closed(
    beamstop_check_devices, sim_run_engine, beamline_parameters
):
    sim_run_engine.add_handler(
        "locate", lambda msg: {"readback": ShutterState.OPEN}, "detector_motion-shutter"
    )
    with pytest.raises(
        RuntimeError,
        match="Unable to proceed with beamstop background check, detector shutter did not close",
    ):
        sim_run_engine.simulate_plan(
            move_beamstop_in_and_verify_using_diode(
                beamstop_check_devices,
                beamline_parameters,
                250,
                800,
            )
        )


@pytest.mark.parametrize(
    "ipin_reading_with_beamstop_out, beamstop_threshold, commissioning_mode, expected_exception",
    [
        [0.099, 0.1, False, BeamObstructedError],
        [0.101, 0.1, False, None],
        [0.099, 0.05, False, None],
        [0.099, 0.15, False, BeamObstructedError],
        [0.099, 0.15, True, None],
    ],
)
@patch(
    "mx_bluesky.common.experiment_plans.beamstop_check._post_beamstop_out_check_actions"
)
@patch("mx_bluesky.common.experiment_plans.beamstop_check.bps.sleep", new=MagicMock())
def test_beamstop_check_checks_beamstop_out_diode_above_threshold_before_second_check(
    mock_post_beamstop_out_actions,
    beamstop_check_devices: BeamstopCheckDevices,
    run_engine: RunEngine,
    beamline_parameters: dict[str, Any],
    ipin_reading_with_beamstop_out,
    beamstop_threshold: float,
    commissioning_mode: bool,
    expected_exception: type[Exception] | None,
):
    set_mock_value(beamstop_check_devices.baton.commissioning, commissioning_mode)
    beamline_parameters["ipin_threshold"] = beamstop_threshold
    value_iter = set_mock_values(
        beamstop_check_devices.ipin.pin_readback, [ipin_reading_with_beamstop_out, 0]
    )
    next(value_iter)

    def bump_reading_iter_on_post_bs_actions(*args, **kwargs):
        next(value_iter)
        yield from _post_beamstop_out_check_actions(*args, **kwargs)

    mock_post_beamstop_out_actions.side_effect = bump_reading_iter_on_post_bs_actions

    with pytest.raises(expected_exception) if expected_exception else nullcontext():
        run_engine(
            move_beamstop_in_and_verify_using_diode(
                beamstop_check_devices,
                beamline_parameters,
                250,
                800,
            )
        )


@pytest.mark.parametrize(
    "ipin_reading_with_beamstop_in, beamstop_threshold, expected_exception",
    [
        [0.101, 0.1, BeamstopNotInPositionError],
        [0.099, 0.1, None],
        [0.051, 0.05, BeamstopNotInPositionError],
        [0.149, 0.15, None],
    ],
)
@patch(
    "mx_bluesky.common.experiment_plans.beamstop_check._post_beamstop_out_check_actions"
)
@patch("mx_bluesky.common.experiment_plans.beamstop_check.bps.sleep", new=MagicMock())
def test_beamstop_check_checks_beamstop_in_diode_below_threshold(
    mock_post_beamstop_out_actions,
    beamstop_check_devices: BeamstopCheckDevices,
    run_engine: RunEngine,
    beamline_parameters: dict[str, Any],
    ipin_reading_with_beamstop_in,
    beamstop_threshold: float,
    expected_exception: type[Exception] | None,
):
    beamline_parameters["ipin_threshold"] = beamstop_threshold
    value_iter = set_mock_values(
        beamstop_check_devices.ipin.pin_readback, [100, ipin_reading_with_beamstop_in]
    )
    next(value_iter)

    def bump_reading_iter_on_post_bs_actions(*args, **kwargs):
        next(value_iter)
        yield from _post_beamstop_out_check_actions(*args, **kwargs)

    mock_post_beamstop_out_actions.side_effect = bump_reading_iter_on_post_bs_actions

    with pytest.raises(expected_exception) if expected_exception else nullcontext():
        run_engine(
            move_beamstop_in_and_verify_using_diode(
                beamstop_check_devices,
                beamline_parameters,
                250,
                800,
            )
        )


def test_beamstop_check_operates_shutter_and_beamstop_during_ipin_check(
    sim_run_engine: RunEngineSimulator,
    beamstop_check_devices: BeamstopCheckDevices,
    beamline_parameters: dict[str, Any],
):
    msgs = sim_run_engine.simulate_plan(
        move_beamstop_in_and_verify_using_diode(
            beamstop_check_devices,
            beamline_parameters,
            250,
            800,
        )
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait"
        and msg.kwargs["group"] == _GROUP_PRE_BEAMSTOP_OUT_CHECK,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is beamstop_check_devices.sample_shutter
        and msg.args[0] == ZebraShutterState.OPEN,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "read"
        and msg.obj is beamstop_check_devices.ipin.pin_readback,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is beamstop_check_devices.sample_shutter
        and msg.args[0] == ZebraShutterState.CLOSE,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is beamstop_check_devices.beamstop.selected_pos
        and msg.args[0] == BeamstopPositions.DATA_COLLECTION
        and msg.kwargs["group"] == _GROUP_POST_BEAMSTOP_OUT_CHECK,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait"
        and msg.kwargs["group"] == _GROUP_POST_BEAMSTOP_OUT_CHECK,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is beamstop_check_devices.sample_shutter
        and msg.args[0] == ZebraShutterState.OPEN,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "read"
        and msg.obj is beamstop_check_devices.ipin.pin_readback,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is beamstop_check_devices.sample_shutter
        and msg.args[0] == ZebraShutterState.CLOSE,
    )


@pytest.mark.parametrize(
    "current_z, min_z, max_z, expected_move",
    [
        [500, 250, 750, None],
        [0, 250, 750, 250],
        [800, 250, 750, 750],
        [250, 300, 750, 300],
        [720, 250, 680, 680],
    ],
)
def test_beamstop_check_moves_detector_if_outside_thresholds(
    beamstop_check_devices,
    sim_run_engine,
    beamline_parameters,
    current_z,
    min_z,
    max_z,
    expected_move,
):
    sim_run_engine.add_handler(
        "locate", lambda msg: {"readback": current_z}, "detector_motion-z"
    )
    msgs = sim_run_engine.simulate_plan(
        move_beamstop_in_and_verify_using_diode(
            beamstop_check_devices,
            beamline_parameters,
            min_z,
            max_z,
        )
    )

    if expected_move is not None:
        assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "set"
            and msg.obj is beamstop_check_devices.detector_motion.z
            and msg.args[0] == expected_move,
        )
    else:
        assert (
            len(
                [
                    msg
                    for msg in msgs
                    if msg.command == "set"
                    and msg.obj is beamstop_check_devices.detector_motion.z
                ]
            )
            == 0
        )



from dodal.devices.zebra.zebra import (
    Zebra,
)
from dodal.devices.zebra.zebra_controlled_shutter import (
    ZebraShutter,
    ZebraShutterControl,
)

from mx_bluesky.hyperion.device_setup_plans.setup_zebra import (
    setup_zebra_for_panda_flyscan,
)


async def _get_shutter_input_2(zebra: Zebra):
    return (
        await zebra.logic_gates.and_gates[zebra.mapping.AND_GATE_FOR_AUTO_SHUTTER]
        .sources[2]
        .get_value()
    )


async def _get_shutter_input_1(zebra: Zebra):
    return (
        await zebra.logic_gates.and_gates[zebra.mapping.AND_GATE_FOR_AUTO_SHUTTER]
        .sources[1]
        .get_value()
    )


async def test_zebra_set_up_for_panda_gridscan(
    run_engine, zebra: Zebra, zebra_shutter: ZebraShutter
):
    run_engine(setup_zebra_for_panda_flyscan(zebra, zebra_shutter, wait=True))
    assert (
        await zebra.output.out_pvs[zebra.mapping.outputs.TTL_DETECTOR].get_value()
        == zebra.mapping.sources.IN1_TTL
    )
    assert (
        await zebra.output.out_pvs[zebra.mapping.outputs.TTL_PANDA].get_value()
        == zebra.mapping.sources.IN3_TTL
    )
    assert await zebra_shutter.control_mode.get_value() == ZebraShutterControl.AUTO
    assert await _get_shutter_input_2(zebra) == zebra.mapping.sources.IN4_TTL
    assert await _get_shutter_input_1(zebra) == zebra.mapping.sources.SOFT_IN1



"""
A minimal test beamline that contains only a baton, for use in tests which need a beamline
but not all the devices, so that test execution isn't slowed down by loading lots of
python modules/creating objects.
See Also:
    use_beamline_t01()
"""

from dodal.device_manager import DeviceManager
from dodal.devices.baton import Baton
from dodal.devices.detector.detector_motion import DetectorMotion
from dodal.devices.synchrotron import Synchrotron
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.utils import BeamlinePrefix, get_beamline_name

BL = get_beamline_name("t01")
PREFIX = BeamlinePrefix(BL)

devices = DeviceManager()


@devices.factory()
def baton() -> Baton:
    return Baton(f"{PREFIX.beamline_prefix}-CS-BATON-01:")


@devices.factory()
def synchrotron() -> Synchrotron:
    return Synchrotron()


@devices.factory()
def xbpm_feedback() -> XBPMFeedback:
    return XBPMFeedback(
        PREFIX.beamline_prefix,
        "xbpm_feedback",
    )


@devices.factory()
def detector_motion() -> DetectorMotion:
    return DetectorMotion(
        device_prefix=f"{PREFIX.beamline_prefix}-MO-DET-01:",
        pmac_prefix=f"{PREFIX.beamline_prefix}-MO-PMAC-02:",
    )



import os
import time

import bluesky.plan_stubs as bps
import pydantic
from bluesky.utils import MsgGenerator
from dodal.common import inject
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.backlight import Backlight
from dodal.devices.beamlines.i04.beam_centre import CentreEllipseMethod
from dodal.devices.beamlines.i04.max_pixel import MaxPixel
from dodal.devices.mx_phase1.beamstop import Beamstop, BeamstopPositions
from dodal.devices.oav.oav_detector import OAV, ZoomControllerWithBeamCentres
from dodal.devices.robot import BartRobot, PinMounted
from dodal.devices.scintillator import InOut, Scintillator
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.devices.zebra.zebra_controlled_shutter import (
    ZebraShutter,
    ZebraShutterControl,
    ZebraShutterState,
)
from ophyd_async.core import InOut as core_INOUT

from mx_bluesky.common.utils.exceptions import BeamlineStateError
from mx_bluesky.common.utils.log import LOGGER

OAV_PREPARE_BEAMLINE_FOR_SCINT_WAIT = "Wait for scint to move in"


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class FindBeamCentresComposite:
    robot: BartRobot
    beamstop: Beamstop
    backlight: Backlight
    scintillator: Scintillator
    xbpm_feedback: XBPMFeedback
    max_pixel: MaxPixel
    centre_ellipse: CentreEllipseMethod
    attenuator: BinaryFilterAttenuator
    zoom_controller: ZoomControllerWithBeamCentres
    shutter: ZebraShutter


def take_oav_image_with_scintillator_in(
    image_name: str | None = None,
    image_path: str = "dls_sw/i04/software/bluesky/scratch",
    transmission: float = 1,
    attenuator: BinaryFilterAttenuator = inject("attenuator"),
    shutter: ZebraShutter = inject("sample_shutter"),
    oav: OAV = inject("oav"),
    robot: BartRobot = inject("robot"),
    beamstop: Beamstop = inject("beamstop"),
    backlight: Backlight = inject("backlight"),
    scintillator: Scintillator = inject("scintillator"),
    xbpm_feedback: XBPMFeedback = inject("xbpm_feedback"),
) -> MsgGenerator:
    """
    Takes an OAV image at specified transmission after necessary checks and preparation steps.

    Args:
        image_name: Name of the OAV image to be saved
        image_path: Path where the image should be saved
        transmission: Transmission of the beam, takes a value from 0 to 1 where
                    1 lets all the beam through and 0 lets none of the beam through.
        devices: These are the specific ophyd-devices used for the plan, the
                    defaults are always correct.
    """

    LOGGER.info("Preparing beamline to take scintillator images...")
    yield from _prepare_beamline_for_scintillator_images(
        robot,
        beamstop,
        backlight,
        scintillator,
        xbpm_feedback,
        shutter,
        OAV_PREPARE_BEAMLINE_FOR_SCINT_WAIT,
    )
    yield from bps.abs_set(
        attenuator, transmission, group=OAV_PREPARE_BEAMLINE_FOR_SCINT_WAIT
    )

    if image_name is None:
        image_name = f"{time.time_ns()}ATT{transmission * 100}"
    LOGGER.info(f"Using image name {image_name}")
    LOGGER.info("Waiting for prepare beamline plan to complete...")
    yield from bps.wait(OAV_PREPARE_BEAMLINE_FOR_SCINT_WAIT)

    LOGGER.info("Opening shutter...")

    yield from bps.abs_set(shutter.control_mode, ZebraShutterControl.MANUAL, wait=True)
    yield from bps.abs_set(shutter, ZebraShutterState.OPEN, wait=True)

    LOGGER.info("Taking image...")

    yield from take_and_save_oav_image(
        file_path=image_path, file_name=image_name, oav=oav
    )


def _prepare_beamline_for_scintillator_images(
    robot: BartRobot,
    beamstop: Beamstop,
    backlight: Backlight,
    scintillator: Scintillator,
    xbpm_feedback: XBPMFeedback,
    shutter: ZebraShutter,
    group: str,
) -> MsgGenerator:
    """
    Prepares the beamline for oav image by making sure the pin is not mounted and
    the beam is on (feedback check). Finally, the scintillator is moved in and the
    shutter opened.
    """
    pin_mounted = yield from bps.rd(robot.gonio_pin_sensor)
    if pin_mounted == PinMounted.PIN_MOUNTED:
        raise BeamlineStateError("Pin should not be mounted!")

    yield from bps.trigger(xbpm_feedback, group=group)

    yield from bps.abs_set(
        beamstop.selected_pos, BeamstopPositions.DATA_COLLECTION, group=group
    )

    yield from bps.abs_set(backlight, core_INOUT.OUT, group=group)

    yield from bps.abs_set(scintillator.selected_pos, InOut.IN, group=group)
    # Not waiting for control mode to be set before opening shutter can result in timeout error
    yield from bps.abs_set(shutter.control_mode, ZebraShutterControl.MANUAL, wait=True)
    yield from bps.abs_set(shutter, ZebraShutterState.OPEN, group=group)


def take_and_save_oav_image(
    file_name: str,
    file_path: str,
    oav: OAV,
) -> MsgGenerator:
    """
    Plan which takes and saves an OAV image to the specified path.
     Args:
        file_name: Filename specifying the name of the image,
        file_path: Path as a string specifying where the image should be saved,
        oav: The OAV to take the image with
    """
    group = "oav image path setting"
    full_file_path = file_path + "/" + file_name
    if not os.path.exists(full_file_path):
        yield from bps.abs_set(oav.snapshot.filename, file_name, group=group)
        yield from bps.abs_set(oav.snapshot.directory, file_path, group=group)
        yield from bps.wait(group, timeout=60)
        yield from bps.trigger(oav.snapshot, wait=True)
    else:
        raise FileExistsError("OAV image file path already exists")


def _max_pixel_at_transmission(
    max_pixel: MaxPixel,
    attenuator: BinaryFilterAttenuator,
    xbpm_feedback: XBPMFeedback,
    transmission: float,
):
    # Potential controls issue on XBPM device means it can mark
    # itself as stable before it really is
    yield from bps.trigger(xbpm_feedback, wait=True)
    yield from bps.mv(attenuator, transmission)
    yield from bps.trigger(max_pixel, wait=True)
    return (yield from bps.rd(max_pixel.max_pixel_val))


def optimise_transmission_with_oav(
    upper_bound: float = 100,
    lower_bound: float = 0,
    target_brightness_fraction: float = 0.75,
    min_transmission_change: float = 5,
    max_iterations: int = 10,
    max_pixel: MaxPixel = inject("max_pixel"),
    attenuator: BinaryFilterAttenuator = inject("attenuator"),
    xbpm_feedback: XBPMFeedback = inject("xbpm_feedback"),
) -> MsgGenerator:
    """
    Plan to find the optimal oav transmission. First the brightest pixel at 100%
    transmission is taken. A fraction of this (target_brightness_fraction) is taken
    as the target - as in the optimal transmission will have it's max pixel as the set
    target. A binary search is used to reach this.
    Args:
        upper_bound: Maximum transmission which will be searched. In percent.
        lower_bound: Minimum transmission which will be searched. In percent.
        target_brightness_fraction: Fraction of the brightest pixel at 100%
                    transmission which should be used as the target max pixel brightness.
        min_transmission_change: If the next search point would require a transmission
                    change less than this then we stop.
        max_iterations: Maximum amount of iterations.
    """

    if upper_bound < lower_bound:
        raise ValueError(
            f"Upper bound ({upper_bound}) must be higher than lower bound {lower_bound}"
        )

    brightest_pixel_at_full_beam = yield from _max_pixel_at_transmission(
        max_pixel, attenuator, xbpm_feedback, 1
    )

    if brightest_pixel_at_full_beam == 0:
        raise ValueError("No beam found at full transmission")

    target_pixel_brightness = brightest_pixel_at_full_beam * target_brightness_fraction
    LOGGER.info(
        f"Optimising until max pixel in image has a value of {target_pixel_brightness}"
    )

    iterations = 0

    while iterations < max_iterations:
        mid = round((upper_bound + lower_bound) / 2, 2)  # limit to 2 dp
        LOGGER.info(f"On iteration {iterations}")

        current_transmission = yield from bps.rd(attenuator.actual_transmission)
        transmission_change_percent = abs(mid - current_transmission * 100)
        if transmission_change_percent < min_transmission_change:
            LOGGER.info(
                f"Next transmission change would be small ({transmission_change_percent}%) so stopping at {current_transmission}"
            )
            return

        brightest_pixel = yield from _max_pixel_at_transmission(
            max_pixel, attenuator, xbpm_feedback, mid / 100
        )

        LOGGER.info(f"Upper bound is: {upper_bound}, Lower bound is: {lower_bound}")
        LOGGER.info(
            f"Testing transmission {mid}, brightest pixel found {brightest_pixel}"
        )

        if target_pixel_brightness == brightest_pixel:
            mid = round(mid, 0)
            LOGGER.info(f"\nOptimal transmission found: {mid}")
            return

        # condition for too low so want to try higher
        elif brightest_pixel < target_pixel_brightness:
            LOGGER.info("Result: Too low \n")
            lower_bound = mid

        # condition for too high so want to try lower
        elif brightest_pixel > target_pixel_brightness:
            LOGGER.info("Result: Too high \n")
            upper_bound = mid
        iterations += 1
    raise StopIteration("Max iterations reached")


def _get_all_zoom_levels(
    zoom_controller: ZoomControllerWithBeamCentres,
) -> MsgGenerator[tuple[str]]:
    zoom_levels = []
    level_signals = [
        centring_device.level_name
        for centring_device in zoom_controller.beam_centres.values()
    ]
    for signal in level_signals:
        level_name = yield from bps.rd(signal)
        if level_name:
            zoom_levels.append(level_name)
    return tuple(zoom_levels)


def find_beam_centres(
    zoom_levels_to_centre: tuple[str, ...] | None = None,
    zoom_levels_to_optimise_transmission: tuple[str, ...] = ("1.0x", "7.5x"),
    composite: FindBeamCentresComposite = inject(""),
) -> MsgGenerator:
    """
    Finds beam centres at the zoom levels given by zoom_levels_to_centre, first
    optimising transmission if the zoom level is in zoom_levels_to_optimise_transmission.

    Note that the previous beam centre values are used to draw an ROI box around the
    OAV image when finding updated beam centre. If the previous values are very wrong,
    this plan may fail or give inaccurate results.

    zoom_levels_to_centre: The levels to do centring at, by default runs at all known
                           zoom levels.
    zoom_levels_to_optimise_transmission: The levels to optimise transmission at,
                           defaults to 1x and 7.5x
    """

    all_zooms = yield from _get_all_zoom_levels(composite.zoom_controller)
    if zoom_levels_to_centre is None:
        zoom_levels_to_centre = all_zooms

    for zoom in [*zoom_levels_to_optimise_transmission, *zoom_levels_to_centre]:
        if zoom not in all_zooms:
            raise ValueError(f"Unknown zoom ({zoom}). Known zooms are {all_zooms}")

    LOGGER.info("Preparing beamline for images...")
    yield from _prepare_beamline_for_scintillator_images(
        composite.robot,
        composite.beamstop,
        composite.backlight,
        composite.scintillator,
        composite.xbpm_feedback,
        composite.shutter,
        OAV_PREPARE_BEAMLINE_FOR_SCINT_WAIT,
    )
    LOGGER.info("Waiting for prepare beamline plan to complete...")
    yield from bps.wait(OAV_PREPARE_BEAMLINE_FOR_SCINT_WAIT)

    for centring_device in composite.zoom_controller.beam_centres.values():
        zoom_name = yield from bps.rd(centring_device.level_name)
        if zoom_name in zoom_levels_to_centre:
            LOGGER.info(f"Moving to zoom level {zoom_name}")
            yield from bps.abs_set(composite.zoom_controller, zoom_name, wait=True)
            if zoom_name in zoom_levels_to_optimise_transmission:
                LOGGER.info(f"Optimising transmission at zoom level {zoom_name}")
                yield from optimise_transmission_with_oav(
                    100,
                    0,
                    max_pixel=composite.max_pixel,
                    attenuator=composite.attenuator,
                    xbpm_feedback=composite.xbpm_feedback,
                )

            yield from bps.trigger(composite.centre_ellipse, wait=True)
            centre_x = yield from bps.rd(composite.centre_ellipse.center_x_val)
            centre_y = yield from bps.rd(composite.centre_ellipse.center_y_val)
            centre_x = round(centre_x)
            centre_y = round(centre_y)
            LOGGER.info(
                f"Writing centre values ({centre_x}, {centre_y}) to OAV PVs at zoom level {zoom_name}"
            )
            yield from bps.mv(
                centring_device.x_centre, centre_x, centring_device.y_centre, centre_y
            )

    LOGGER.info("Find beam centre plan completed!")


def find_and_set_beam_centre_at_current_zoom_and_transmission(
    composite: FindBeamCentresComposite = inject(""),
):
    """Finds the beam centre at the current zoom level."""
    current_zoom_level = yield from bps.rd(composite.zoom_controller.level)
    yield from find_beam_centres(
        zoom_levels_to_centre=(current_zoom_level,),
        zoom_levels_to_optimise_transmission=(),
        composite=composite,
    )



from collections.abc import Sequence

from blueapi.core.context import BlueskyContext
from bluesky import plan_stubs as bps
from bluesky import preprocessors as bpp
from bluesky.utils import MsgGenerator, RunEngineInterrupted
from dodal.common.beamlines.commissioning_mode import set_commissioning_signal
from dodal.devices.baton import Baton
from dodal.devices.synchrotron import Synchrotron
from pydantic import BaseModel

from mx_bluesky.common.external_interaction.alerting import (
    AlertService,
    get_alerting_service,
)
from mx_bluesky.common.utils.context import (
    find_device_in_context,
)
from mx_bluesky.common.utils.exceptions import BeamlineCheckFailureError
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion._plan_runner_params import (
    RobotUnload,
    UDCCleanup,
    UDCDefaultState,
)
from mx_bluesky.hyperion.external_interaction.agamemnon import (
    create_parameters_from_agamemnon,
)
from mx_bluesky.hyperion.external_interaction.alerting.constants import Subjects
from mx_bluesky.hyperion.plan_runner import PlanError, PlanRunner
from mx_bluesky.hyperion.utils.context import (
    clear_all_device_caches,
    setup_devices,
)

HYPERION_USER = "Hyperion"
NO_USER = "None"
COUNTDOWN_THRESHOLD_SECONDS = 600


def run_forever(runner: PlanRunner):
    try:
        while True:
            try:
                run_udc_when_requested(runner.context, runner)
            except PlanError as e:
                LOGGER.info(
                    "Caught exception during plan execution, stopped and waiting for baton.",
                    exc_info=e,
                )

    except RunEngineInterrupted:
        # In the event that BlueskyRunner.stop() or shutdown() was called then
        # RunEngine.abort() will have been called and we will get RunEngineInterrupted
        LOGGER.info(
            f"RunEngine was interrupted. Runner state is {runner.current_status}, "
            f"run engine is {runner.run_engine.state}"
        )


def run_udc_when_requested(context: BlueskyContext, runner: PlanRunner):
    """This will wait for the baton to be handed to hyperion and then run through the
    UDC queue from agamemnon until:
      1. There are no more instructions from agamemnon
      2. There is an error on the beamline
      3. The baton is requested by another party
      4. A shutdown is requested

    In the case of 1. 2. or 4. hyperion will immediately release the baton. In the case of
    3. the baton will be released after the next collection has finished."""

    baton = _get_baton(context)

    def acquire_baton() -> MsgGenerator:
        yield from _wait_for_hyperion_requested(baton)
        LOGGER.debug("Hyperion is now current baton holder.")
        yield from bps.abs_set(baton.current_user, HYPERION_USER)

    def collect() -> MsgGenerator:
        """
        Move to the default state for collection, then enter a loop fetching instructions
        from Agamemnon and continue the loop until any of the following occur:
        * A user requests the baton away from Hyperion
        * Hyperion releases the baton when Agamemnon has no more instructions
        * The RunEngine raises a RequestAbort exception, most likely due to a shutdown command
        * A plan raises an exception not of type WarningError (which is then wrapped as a PlanError)
        Args:
            baton: The baton device
            runner: The runner
        """

        _raise_udc_start_alert(get_alerting_service())
        yield from bpp.contingency_wrapper(
            runner.decode_and_execute(None, [UDCDefaultState()]),
            except_plan=trap_default_state_exception,
            auto_raise=False,
        )

        # re-fetch the baton because the device has been reinstantiated
        baton = _get_baton(context)
        current_visit: str | None = None
        while (yield from _is_requesting_baton(baton)):
            abort = yield from _abort_if_countdown_too_low(context, baton)
            if abort:
                break
            current_visit = yield from _fetch_and_process_agamemnon_instruction(
                baton, runner, current_visit
            )
        yield from runner.decode_and_execute(current_visit, [RobotUnload()])

    def clean_up_and_release_baton() -> MsgGenerator:
        # If hyperion has given up the baton itself we need to also release requested
        # user so that hyperion doesn't think we're requested again
        baton = _get_baton(context)
        yield from runner.decode_and_execute(None, [UDCCleanup()])
        previous_requested_user = yield from _unrequest_baton(baton)
        LOGGER.debug("Hyperion no longer current baton holder.")
        yield from bps.abs_set(baton.current_user, NO_USER, wait=True)
        _raise_baton_released_alert(get_alerting_service(), previous_requested_user)

    def trap_default_state_exception(e: Exception):
        yield from bps.null()
        if isinstance(e, BeamlineCheckFailureError):
            LOGGER.warning("Caught default state check failure:", exc_info=e)
            raise PlanError("Caught default state check failure") from e
        else:
            LOGGER.warning("Caught unexpected exception", exc_info=e)
            raise PlanError("Unexpected exception from UDC Default State plan") from e

    def collect_then_release() -> MsgGenerator:
        yield from bpp.contingency_wrapper(
            collect(),
            final_plan=clean_up_and_release_baton,
        )

    context.run_engine(acquire_baton())
    _initialise_udc(context, runner.is_dev_mode)
    context.run_engine(collect_then_release())


def _initialise_udc(context: BlueskyContext, dev_mode: bool):
    """
    Perform all initialisation that happens at the start of UDC just after the
    baton is acquired, but before we execute any plans or move hardware.

    Beamline devices are unloaded and reloaded in order to pick up any new configuration,
    bluesky context gets new set of devices.
    """
    LOGGER.info("Initialising mx-bluesky for UDC start...")
    clear_all_device_caches(context)
    LOGGER.debug("Reinitialising beamline devices")
    setup_devices(context, dev_mode)
    set_commissioning_signal(_get_baton(context).commissioning)


def _wait_for_hyperion_requested(baton: Baton):
    LOGGER.debug("Hyperion waiting for baton...")
    sleep_per_check = 0.1
    while True:
        requested_user = yield from bps.rd(baton.requested_user)
        if requested_user == HYPERION_USER:
            LOGGER.debug("Baton requested for Hyperion")
            break
        yield from bps.sleep(sleep_per_check)


def _fetch_and_process_agamemnon_instruction(
    baton: Baton, runner: PlanRunner, current_visit: str | None
) -> MsgGenerator[str | None]:
    parameter_list: Sequence[BaseModel] = create_parameters_from_agamemnon()
    if parameter_list:
        current_visit = yield from runner.decode_and_execute(
            current_visit, parameter_list
        )
    else:
        yield from _release_baton_on_completed_alert(baton)
    return current_visit


def _raise_udc_start_alert(alert_service: AlertService):
    alert_service.raise_alert(
        Subjects.UDC_STARTED, "Unattended Data Collection has started.", {}
    )


def _raise_baton_released_alert(alert_service: AlertService, baton_requester: str):
    alert_service.raise_alert(
        Subjects.UDC_BATON_RELEASED,
        f"Hyperion has released the baton. The baton is currently requested by:"
        f" {baton_requester}",
        {},
    )


def _raise_udc_completed_alert(alert_service: AlertService):
    alert_service.raise_alert(
        Subjects.UDC_COMPLETED,
        "Hyperion UDC has completed all pending Agamemnon requests.",
        {},
    )


def _is_requesting_baton(baton: Baton) -> MsgGenerator:
    requested_user = yield from bps.rd(baton.requested_user)
    return requested_user == HYPERION_USER


def _get_baton(context: BlueskyContext) -> Baton:
    return find_device_in_context(context, "baton", Baton)


def _get_synchrotron(context: BlueskyContext) -> Synchrotron:
    return find_device_in_context(context, "synchrotron", Synchrotron)


def _unrequest_baton(baton: Baton) -> MsgGenerator[str]:
    """Relinquish the requested user of the baton if it is not already requested
    by another user.

    Returns:
        The previously requested user, or NO_USER if no user was already requested.
    """
    requested_user = yield from bps.rd(baton.requested_user)
    if requested_user == HYPERION_USER:
        LOGGER.debug("Hyperion no longer requesting baton")
        yield from bps.abs_set(baton.requested_user, NO_USER)
        return NO_USER
    return requested_user


def _abort_if_countdown_too_low(
    context: BlueskyContext, baton: Baton
) -> MsgGenerator[bool]:
    synchrotron = _get_synchrotron(context)
    countdown = yield from bps.rd(synchrotron.machine_user_countdown)

    LOGGER.info(f"Synchrotron beam countdown is {countdown} seconds")

    if countdown < COUNTDOWN_THRESHOLD_SECONDS:
        LOGGER.info("Synchrotron machine countdown too low")
        yield from _release_baton_on_completed_alert(baton)
        return True

    return False


def _release_baton_on_completed_alert(baton) -> MsgGenerator:
    _raise_udc_completed_alert(get_alerting_service())
    yield from _unrequest_baton(baton)



import asyncio
from functools import partial
from pathlib import Path
from unittest.mock import AsyncMock, MagicMock, call, patch

import bluesky.plan_stubs as bps
from bluesky.preprocessors import run_decorator
from bluesky.run_engine import RunEngine
from dodal.devices.beamlines.i24.commissioning_jungfrau import CommissioningJungfrau
from ophyd_async.core import (
    TriggerInfo,
    completed_status,
    set_mock_value,
)
from ophyd_async.fastcs.jungfrau import GainMode

from mx_bluesky.beamlines.i24.jungfrau_commissioning.plan_stubs.plan_utils import (
    JF_COMPLETE_GROUP,
    fly_jungfrau,
)


async def test_fly_jungfrau(
    run_engine: RunEngine, jungfrau: CommissioningJungfrau, tmp_path: Path
):
    set_mock_value(jungfrau._writer.frame_counter, 10)
    mock_stop = AsyncMock()
    jungfrau.drv.acquisition_stop.trigger = mock_stop

    filename = "test"

    @run_decorator(md={"detector_file_template": filename})
    def _open_run_and_fly():
        frames = 5
        status = yield from fly_jungfrau(
            jungfrau,
            TriggerInfo(livetime=1e-3, exposures_per_event=frames),
            GainMode.DYNAMIC,
        )
        val = 0
        while not status.done:
            val += 1
            set_mock_value(jungfrau._writer.frame_counter, val)
            yield from bps.sleep(0.001)
        yield from bps.wait(JF_COMPLETE_GROUP)
        assert val == frames
        assert (
            yield from bps.rd(jungfrau._writer.file_path)
        ) == f"{tmp_path}/0000_{filename}"

    run_engine(_open_run_and_fly())
    await asyncio.sleep(0)


@patch(
    "mx_bluesky.beamlines.i24.jungfrau_commissioning.plan_stubs.plan_utils.log_on_percentage_complete",
    new=MagicMock(),
)
async def test_fly_jungfrau_does_read_plan_after_prepare(
    run_engine: RunEngine, jungfrau: CommissioningJungfrau
):
    mock_stop = AsyncMock()
    jungfrau.drv.acquisition_stop.trigger = mock_stop

    read_hardware = MagicMock()

    filename = "test"
    jungfrau.prepare = MagicMock(side_effect=lambda _: completed_status())

    parent_mock = MagicMock()
    parent_mock.attach_mock(jungfrau.prepare, "jungfrau_prepare")
    parent_mock.attach_mock(read_hardware, "read_hardware")
    jungfrau.kickoff = MagicMock(side_effect=lambda: completed_status())
    jungfrau.complete = MagicMock(side_effect=lambda: completed_status())
    test_trigger_info = TriggerInfo(livetime=1e-3, exposures_per_event=5)

    @run_decorator(md={"detector_file_template": filename})
    def fly_in_run():
        yield from fly_jungfrau(
            jungfrau,
            test_trigger_info,
            GainMode.DYNAMIC,
            read_hardware_after_prepare_plan=partial(read_hardware),
        )

    run_engine(fly_in_run())
    assert parent_mock.method_calls == [
        call.jungfrau_prepare(test_trigger_info),
        call.read_hardware(),
    ]



import pytest

from mx_bluesky.common.parameters.rotation import (
    RotationScan,
)
from mx_bluesky.hyperion.parameters.constants import CONST

_UID_ROTATION_OUTER = "d8bee3ee-f614-4e7a-a516-25d6b9e87ef3"


@pytest.fixture
def test_rotation_start_outer_document(dummy_rotation_params: RotationScan):
    dummy_single_rotation_params = next(dummy_rotation_params.single_rotation_scans)
    return {
        "uid": _UID_ROTATION_OUTER,
        "subplan_name": CONST.PLAN.ROTATION_OUTER,
        "mx_bluesky_parameters": dummy_single_rotation_params.model_dump_json(),
    }


@pytest.fixture
def test_rotation_stop_outer_document():
    return {
        "run_start": _UID_ROTATION_OUTER,
        "time": 1666604300.0310638,
        "uid": "65b2bde5-5740-42d7-9047-e860e06fbe15",
        "exit_status": "Success",
        "reason": "",
    }



import os
from collections.abc import Sequence
from functools import partial
from pathlib import Path
from unittest.mock import ANY, MagicMock, Mock, call, patch

import bluesky.plan_stubs as bps
import pytest
from _pytest.fixtures import FixtureRequest
from bluesky.preprocessors import run_decorator, set_run_key_decorator
from bluesky.run_engine import RunEngine
from dodal.devices.areadetector.plugins.mjpg import MJPG
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.smargon import Smargon
from ophyd_async.core import AsyncStatus, set_mock_value
from PIL import Image

from mx_bluesky.common.parameters.components import WithSnapshot
from mx_bluesky.common.parameters.constants import DocDescriptorNames
from mx_bluesky.common.parameters.rotation import (
    SingleRotationScan,
)
from mx_bluesky.hyperion.external_interaction.callbacks.snapshot_callback import (
    BeamDrawingCallback,
)
from mx_bluesky.hyperion.parameters.constants import CONST

from ......conftest import assert_images_pixelwise_equal, raw_params_from_file


@pytest.fixture
def params_take_snapshots(tmp_path):
    return SingleRotationScan(
        **raw_params_from_file(
            "tests/test_data/parameter_json_files/good_test_rotation_scan_parameters.json",
            tmp_path,
        )
    )


@pytest.fixture
def next_snapshot():
    return MagicMock(
        return_value="tests/test_data/test_images/generate_snapshot_input.png"
    )


@pytest.fixture
def oav_with_snapshots(oav: OAV, next_snapshot):
    @AsyncStatus.wrap
    async def fake_trigger(mjpg: MJPG):
        with Image.open(next_snapshot()) as image:
            # don't do full post-processing to save on slow PIL image save calls
            await mjpg._save_image(image)

    oav.snapshot.trigger = MagicMock(side_effect=partial(fake_trigger, oav.snapshot))
    oav.grid_snapshot.trigger = MagicMock(
        side_effect=partial(fake_trigger, oav.grid_snapshot)
    )
    yield oav


@pytest.fixture(autouse=True)
def optimise_pil_for_speed():
    with patch(
        "mx_bluesky.hyperion.external_interaction.callbacks.snapshot_callback.COMPRESSION_LEVEL",
        1,
    ):
        yield


def simple_rotation_snapshot_plan(
    oav: OAV, snapshot_directory: Path, params: WithSnapshot
):
    @set_run_key_decorator(CONST.PLAN.LOAD_CENTRE_COLLECT)
    @run_decorator(
        md={
            "activate_callbacks": ["BeamDrawingCallback"],
            "with_snapshot": params.model_dump_json(),
        }
    )
    def inner():
        yield from bps.abs_set(oav.snapshot.directory, str(snapshot_directory))
        yield from bps.abs_set(oav.snapshot.filename, "test_filename")
        yield from bps.trigger(oav.snapshot, wait=True)
        yield from bps.create(DocDescriptorNames.OAV_ROTATION_SNAPSHOT_TRIGGERED)
        yield from bps.read(oav)
        yield from bps.save()

    yield from inner()


def simple_take_grid_snapshot_and_generate_rotation_snapshot_plan(
    oav: OAV,
    smargon: Smargon,
    snapshot_directory: Path,
    chis: Sequence[int] = [0],
    grid_smargon_mm: tuple[float, float, float] = (-0.614, 0.0259, 0.250),
    rotation_smargon_mm: tuple[float, float, float] = (-0.4634, 0.0187, 0.2482),
):
    @set_run_key_decorator(CONST.PLAN.ROTATION_MAIN)
    @run_decorator(
        md={
            "subplan_name": CONST.PLAN.ROTATION_MAIN,
        }
    )
    def rotation_plan(rotation_snapshot_dir: Path, chi: float):
        yield from bps.abs_set(oav.snapshot.directory, str(rotation_snapshot_dir))
        for _omega in (
            0,
            270,
        ):
            yield from bps.create(DocDescriptorNames.OAV_ROTATION_SNAPSHOT_TRIGGERED)
            yield from bps.read(oav)  # Capture path info for generated snapshot
            yield from bps.read(smargon)  # Capture the current sample x, y, z
            yield from bps.save()

    @set_run_key_decorator(CONST.PLAN.LOAD_CENTRE_COLLECT)
    @run_decorator(
        md={
            "activate_callbacks": ["BeamDrawingCallback"],
            "with_snapshot": WithSnapshot.model_validate(
                {
                    "snapshot_directory": snapshot_directory,
                    "use_grid_snapshots": True,
                }
            ).model_dump_json(),
        }
    )
    def inner():
        grid_snapshot_dir = snapshot_directory / "grid_snapshots"
        rotation_snapshot_dir = snapshot_directory / "rotation_snapshots"
        os.mkdir(grid_snapshot_dir)
        yield from bps.mv(
            smargon.x,
            grid_smargon_mm[0],
            smargon.y,
            grid_smargon_mm[1],
            smargon.z,
            grid_smargon_mm[2],
        )
        for omega in (
            0,
            -90,
        ):
            yield from bps.abs_set(smargon.omega, omega, wait=True)
            yield from bps.abs_set(
                oav.grid_snapshot.directory, str(grid_snapshot_dir), wait=True
            )
            yield from bps.abs_set(
                oav.grid_snapshot.filename,
                f"my_grid_snapshot_prefix_{omega}",
                wait=True,
            )
            yield from bps.trigger(oav.grid_snapshot, wait=True)
            yield from bps.create(DocDescriptorNames.OAV_GRID_SNAPSHOT_TRIGGERED)
            yield from bps.read(oav)  # Capture base image path
            yield from bps.read(smargon)  # Capture base image sample x, y, z, omega
            yield from bps.save()
        yield from bps.mv(
            smargon.x,
            rotation_smargon_mm[0],
            smargon.y,
            rotation_smargon_mm[1],
            smargon.z,
            rotation_smargon_mm[2],
        )
        yield from bps.wait()
        for chi in chis:
            yield from rotation_plan(rotation_snapshot_dir, chi)

    yield from inner()


class TestGeneratedSnapshots:
    @pytest.fixture()
    def next_snapshot(self, request: FixtureRequest):
        return MagicMock(side_effect=request.param)

    @pytest.fixture()
    def test_config_files(self):
        return {
            "zoom_params_file": "tests/test_data/test_jCameraManZoomLevels.xml",
            "oav_config_json": "tests/test_data/test_daq_configuration/OAVCentring_hyperion.json",
            "display_config": "tests/test_data/test_daq_configuration/display.configuration",
        }

    @pytest.fixture()
    def snapshot_oav_with_1x_zoom(self, oav_with_snapshots):
        set_mock_value(oav_with_snapshots.zoom_controller.level, "1.0x")
        return oav_with_snapshots

    @pytest.mark.parametrize("chis", [[0], [0, 30]])
    @pytest.mark.parametrize(
        "next_snapshot, expected_0_img, expected_270_img, grid_smargon_mm, "
        "rotation_smargon_mm",
        [
            [
                [
                    "tests/test_data/test_images/thau_1_91_0.png",
                    "tests/test_data/test_images/thau_1_91_90.png",
                ],
                "thau_1_91_expected_0.png",
                "thau_1_91_expected_270.png",
                (-0.614, 0.0259, 0.250),
                (-0.4634, 0.0187, 0.2482),
            ],
            [
                [
                    "tests/test_data/test_images/thau_1_91_0.png",
                    "tests/test_data/test_images/thau_1_91_90.png",
                ],
                "thau_1_91_expected_0.png",
                "thau_1_91_expected_270.png",
                (-0.614, 0.0259, 0.250),
                (-0.4634, 0.0187, 0.2482),
            ],
            [
                [
                    "tests/test_data/test_images/ins_15_33_0.png",
                    "tests/test_data/test_images/ins_15_33_90.png",
                ],
                "ins_15_33_expected_0.png",
                "ins_15_33_expected_270.png",
                (0.4678, -0.5481, -0.3128),
                (0.335, -0.532, -0.243),
            ],
        ],
        indirect=["next_snapshot"],
    )
    def test_snapshot_callback_generate_snapshot_from_gridscan(
        self,
        tmp_path: Path,
        run_engine: RunEngine,
        snapshot_oav_with_1x_zoom: OAV,
        smargon: Smargon,
        chis: Sequence[int],
        expected_0_img: str,
        expected_270_img: str,
        grid_smargon_mm: tuple[float, float, float],
        rotation_smargon_mm: tuple[float, float, float],
    ):
        downstream_cb = Mock()
        callback = BeamDrawingCallback(emit=downstream_cb)

        run_engine.subscribe(callback)
        run_engine(
            simple_take_grid_snapshot_and_generate_rotation_snapshot_plan(
                snapshot_oav_with_1x_zoom,
                smargon,
                tmp_path,
                chis,
                grid_smargon_mm,
                rotation_smargon_mm,
            )
        )

        downstream_calls = downstream_cb.mock_calls
        descriptors_to_event_names = {
            c.args[1]["uid"]: c.args[1]["name"]
            for c in downstream_calls
            if c.args[0] == "descriptor"
        }
        rotation_snapshot_events = [
            c.args[1]["data"]
            for c in downstream_calls
            if c.args[0] == "event"
            and descriptors_to_event_names.get(c.args[1]["descriptor"])
            == DocDescriptorNames.OAV_ROTATION_SNAPSHOT_TRIGGERED
        ]

        i = 0
        for _chi in chis:
            for _omega, expected_image_path in zip(
                [0, 270],
                [
                    f"tests/test_data/test_images/{expected_0_img}",
                    f"tests/test_data/test_images/{expected_270_img}",
                ],
                strict=False,
            ):
                generated_image_path = rotation_snapshot_events[i][
                    "oav-snapshot-last_saved_path"
                ]
                assert_images_pixelwise_equal(
                    generated_image_path,
                    expected_image_path,
                )
                i += 1

    @pytest.mark.parametrize(
        "next_snapshot",
        [
            [
                "tests/test_data/test_images/thau_1_91_0.png",
                "tests/test_data/test_images/thau_1_91_90.png",
            ]
        ],
        indirect=True,
    )
    @pytest.mark.parametrize(
        "expected_px_1, expected_px_2, grid_smargon_mm, rotation_smargon_mm",
        [
            [(397.0, 373.0), (397.0, 373.0), (0, 0, 0), (0, 0, 0)],
            [(497.0, 373.0), (497.0, 373.0), (0, 0, 0), (0.287, 0, 0)],
            [(397.0, 473.0), (397.0, 373.0), (0, 0, 0), (0, 0.287, 0)],
            [(397.0, 373.0), (397.0, 473.0), (0, 0, 0), (0, 0, 0.287)],
        ],
    )
    @patch(
        "mx_bluesky.hyperion.external_interaction.callbacks.snapshot_callback.draw_crosshair"
    )
    def test_draw_crosshair_in_expected_position(
        self,
        mock_draw_crosshair,
        snapshot_oav_with_1x_zoom,
        smargon,
        tmp_path,
        run_engine,
        expected_px_1,
        expected_px_2,
        grid_smargon_mm,
        rotation_smargon_mm,
    ):
        callback = BeamDrawingCallback()

        run_engine.subscribe(callback)
        run_engine(
            simple_take_grid_snapshot_and_generate_rotation_snapshot_plan(
                snapshot_oav_with_1x_zoom,
                smargon,
                tmp_path,
                [0],
                grid_smargon_mm,
                rotation_smargon_mm,
            )
        )

        mock_draw_crosshair.assert_has_calls(
            [
                call(ANY, expected_px_1[0], expected_px_1[1]),
                call(ANY, expected_px_2[0], expected_px_2[1]),
            ]
        )


def test_snapshot_callback_loads_and_saves_updated_snapshot_propagates_event(
    tmp_path: Path,
    run_engine: RunEngine,
    oav_with_snapshots: OAV,
    params_take_snapshots: SingleRotationScan,
):
    oav = oav_with_snapshots
    downstream_cb = Mock()
    callback = BeamDrawingCallback(emit=downstream_cb)

    run_engine.subscribe(callback)
    run_engine(simple_rotation_snapshot_plan(oav, tmp_path, params_take_snapshots))

    generated_image_path = str(tmp_path / "test_filename_with_beam_centre.png")
    assert_images_pixelwise_equal(
        generated_image_path, "tests/test_data/test_images/generate_snapshot_output.png"
    )

    downstream_calls = downstream_cb.mock_calls
    assert downstream_calls[0].args[0] == "start"
    assert downstream_calls[1].args[0] == "descriptor"
    assert downstream_calls[2].args[0] == "event"
    assert (
        downstream_calls[2].args[1]["data"]["oav-snapshot-last_saved_path"]
        == generated_image_path
    )
    assert downstream_calls[3].args[0] == "stop"



import pytest
from dodal.beamlines.i24 import VerticalGoniometer
from dodal.devices.attenuator.attenuator import EnumFilterAttenuator
from dodal.devices.beamlines.i24.aperture import Aperture
from dodal.devices.beamlines.i24.beamstop import Beamstop
from dodal.devices.beamlines.i24.commissioning_jungfrau import CommissioningJungfrau
from dodal.devices.beamlines.i24.dcm import DCM
from dodal.devices.beamlines.i24.dual_backlight import DualBacklight
from dodal.devices.hutch_shutter import HutchInterlock, InterlockedHutchShutter
from dodal.devices.motors import YZStage
from dodal.devices.synchrotron import Synchrotron
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.devices.zebra.zebra import Zebra
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutter
from ophyd_async.core import init_devices

from mx_bluesky.beamlines.i24.jungfrau_commissioning.experiment_plans.rotation_scan_plan import (
    RotationScanComposite,
)


@pytest.fixture
def rotation_composite(
    jungfrau: CommissioningJungfrau, zebra: Zebra, enum_attenuator: EnumFilterAttenuator
) -> RotationScanComposite:
    with init_devices(mock=True):
        aperture = Aperture("")
        gonio = VerticalGoniometer("")
        synchrotron = Synchrotron("")
        sample_shutter = ZebraShutter("")
        xbpm_feedback = XBPMFeedback("")
        hutch_shutter = InterlockedHutchShutter("", HutchInterlock(""))
        beamstop = Beamstop("")
        det_stage = YZStage(
            "",
            name="detector_motion",  # Name of device in i24 dodal module
        )
        backlight = DualBacklight("")
        dcm = DCM("", "")

    composite = RotationScanComposite(
        aperture,
        enum_attenuator,
        jungfrau,
        gonio,
        synchrotron,
        sample_shutter,
        zebra,
        xbpm_feedback,
        hutch_shutter,
        beamstop,
        det_stage,
        backlight,
        dcm,
    )

    return composite


@pytest.fixture(autouse=True)
def always_use_i24_beamline(use_i24_beamline): ...



import os
from enum import Enum, StrEnum

from dodal.devices.aperturescatterguard import ApertureValue
from dodal.devices.detector import EIGER2_X_16M_SIZE
from dodal.devices.zocalo.zocalo_constants import ZOCALO_ENV as ZOCALO_ENV_FROM_DODAL
from dodal.utils import get_beamline_name
from pydantic.dataclasses import dataclass

from mx_bluesky.definitions import ROOT_DIR

# Use as visit if numtracker is being used
USE_NUMTRACKER = "from numtracker"

BEAMLINE = get_beamline_name("test")
TEST_MODE = BEAMLINE == "test"
ZEBRA_STATUS_TIMEOUT = 30

GDA_DOMAIN_PROPERTIES_PATH = (
    "tests/test_data/test_domain_properties"
    if TEST_MODE
    else (f"/dls_sw/{BEAMLINE}/software/daq_configuration/domain/domain.properties")
)


@dataclass(frozen=True)
class DocDescriptorNames:
    # Robot load/unload event descriptor
    ROBOT_PRE_LOAD = "robot_update_pre_load"
    ROBOT_UPDATE = "robot_update"
    # For callbacks to use
    OAV_ROTATION_SNAPSHOT_TRIGGERED = "rotation_snapshot_triggered"
    OAV_GRID_SNAPSHOT_TRIGGERED = "snapshot_to_ispyb"
    HARDWARE_READ_PRE = "read_hardware_for_callbacks_pre_collection"
    HARDWARE_READ_DURING = "read_hardware_for_callbacks_during_collection"
    SAMPLE_HANDLING_EXCEPTION = "sample_handling_exception"
    ZOCALO_HW_READ = "zocalo_read_hardware_plan"
    FLYSCAN_RESULTS = "flyscan_results_obtained"


def _get_oav_config_json_path():
    if TEST_MODE:
        return "tests/test_data/test_OAVCentring.json"
    elif BEAMLINE == "i03":
        return f"/dls_sw/{BEAMLINE}/software/daq_configuration/json/OAVCentring_hyperion.json"
    elif BEAMLINE == "aithre":
        return "/dls/science/groups/i23/aithre/daq_configuration/json/OAVCentring_aithre.json"
    else:
        return f"/dls_sw/{BEAMLINE}/software/daq_configuration/json/OAVCentring.json"


@dataclass(frozen=True)
class OavConstants:
    OAV_CONFIG_JSON = _get_oav_config_json_path()


@dataclass(frozen=True)
class PlanNameConstants:
    """
    See https://diamondlightsource.github.io/mx-bluesky/main/developer/general/explanations/callback_and_run_logic.html
    for how these are used.
    """

    # UDC plan
    LOAD_CENTRE_COLLECT = "load_centre_collect"
    # Robot subplans
    ROBOT_LOAD = "robot_load"
    ROBOT_UNLOAD = "robot_unload"
    # Gridscan and xrc subplans
    GRID_DETECT_AND_DO_GRIDSCAN = "grid_detect_and_do_gridscan"
    GRIDSCAN_OUTER = "run_gridscan_move_and_tidy"
    DO_FGS = "do_fgs"
    FLYSCAN_RESULTS = "xray_centre_results"
    # Rotation scan
    ROTATION_MULTI = "multi_rotation_wrapper"
    ROTATION_MULTI_OUTER = "multi_rotation_outer"
    ROTATION_OUTER = "rotation_scan_with_cleanup"
    ROTATION_MAIN = "rotation_scan_main"

    SET_ENERGY = "set_energy"


@dataclass(frozen=True)
class EnvironmentConstants:
    ZOCALO_ENV = ZOCALO_ENV_FROM_DODAL


@dataclass(frozen=True)
class HardwareConstants:
    OAV_REFRESH_DELAY = 0.3
    PANDA_FGS_RUN_UP_DEFAULT = 0.17
    CRYOJET_MARGIN_MM = 0.2
    TIP_OFFSET_UM = 0
    MAX_CRYO_TEMP_K = 110
    MAX_CRYO_PRESSURE_BAR = 0.1

    # Value quoted in https://www.dectris.com/en/detectors/x-ray-detectors/eiger2/eiger2-for-synchrotrons/eiger2-x/,
    # causes dropped frames, so increase value for safety
    PANDA_FGS_EIGER_DEADTIME_S = 5e-5


@dataclass(frozen=True)
class GridscanParamConstants:
    WIDTH_UM = 600.0
    EXPOSURE_TIME_S = 0.004
    USE_ROI = True
    BOX_WIDTH_UM = 20.0
    OMEGA_1 = 0.0
    OMEGA_2 = 90.0
    PANDA_RUN_UP_DISTANCE_MM = 0.2
    ZOCALO_MIN_TOTAL_COUNT_THRESHOLD = 3


@dataclass(frozen=True)
class RotationParamConstants:
    DEFAULT_APERTURE_POSITION = ApertureValue.LARGE
    DEFAULT_SHUTTER_TIME_S = 0.06
    OMEGA_FLIP = True  # See https://github.com/DiamondLightSource/mx-bluesky/issues/1223 to make beamline-specific


@dataclass(frozen=True)
class DetectorParamConstants:
    BEAM_XY_LUT_PATH = (
        "tests/test_data/test_det_dist_converter.txt"
        if TEST_MODE
        else f"/dls_sw/{BEAMLINE}/software/daq_configuration/lookup/DetDistToBeamXYConverter.txt"
    )
    DETECTOR = EIGER2_X_16M_SIZE


@dataclass(frozen=True)
class ExperimentParamConstants:
    DETECTOR = DetectorParamConstants()
    GRIDSCAN = GridscanParamConstants()
    ROTATION = RotationParamConstants()


@dataclass(frozen=True)
class PlanGroupCheckpointConstants:
    # For places to synchronise / stop and wait in plans, use as bluesky group names
    GRID_READY_FOR_DC = "grid_ready_for_data_collection"
    ROTATION_READY_FOR_DC = "rotation_ready_for_data_collection"
    MOVE_GONIO_TO_START = "move_gonio_to_start"
    READY_FOR_OAV = "ready_for_oav"
    PREPARE_APERTURE = "prepare_aperture"
    SETUP_ZEBRA_FOR_ROTATION = "setup_zebra_for_rotation"


# Eventually replace below with https://github.com/DiamondLightSource/mx-bluesky/issues/798
@dataclass(frozen=True)
class DeviceSettingsConstants:
    PANDA_FLYSCAN_SETTINGS_FILENAME = "panda-gridscan"
    PANDA_FLYSCAN_SETTINGS_DIR = os.path.abspath(
        f"{ROOT_DIR}/hyperion/resources/panda/"
    )


class Actions(Enum):
    START = "start"
    STOP = "stop"
    SHUTDOWN = "shutdown"
    STATUS = "status"


class Status(Enum):
    WARN = "Warn"
    FAILED = "Failed"
    SUCCESS = "Success"
    BUSY = "Busy"
    ABORTING = "Aborting"
    IDLE = "Idle"


@dataclass
class FeatureSettings: ...  # List of features and their default values. Subclasses must also be a pydantic dataclass


class FeatureSettingSources(
    StrEnum
): ...  # List of features and the name of that property in domain.properties



from mx_bluesky.common.external_interaction.alerting._service import (
    AlertService,
    Metadata,
    get_alerting_service,
    set_alerting_service,
)

__all__ = [
    "AlertService",
    "Metadata",
    "get_alerting_service",
    "set_alerting_service",
]



from __future__ import annotations

import pydantic
from dodal.devices.attenuator.attenuator import EnumFilterAttenuator
from dodal.devices.beamlines.i24.aperture import Aperture
from dodal.devices.beamlines.i24.beamstop import Beamstop
from dodal.devices.beamlines.i24.commissioning_jungfrau import CommissioningJungfrau
from dodal.devices.beamlines.i24.dcm import DCM
from dodal.devices.beamlines.i24.dual_backlight import DualBacklight
from dodal.devices.beamlines.i24.vgonio import VerticalGoniometer
from dodal.devices.hutch_shutter import InterlockedHutchShutter
from dodal.devices.motors import YZStage
from dodal.devices.synchrotron import Synchrotron
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.devices.zebra.zebra import Zebra
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutter


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class RotationScanComposite:
    """All devices which are directly or indirectly required by this plan"""

    aperture: Aperture
    attenuator: EnumFilterAttenuator
    jungfrau: CommissioningJungfrau
    gonio: VerticalGoniometer
    synchrotron: Synchrotron
    sample_shutter: ZebraShutter
    zebra: Zebra
    xbpm_feedback: XBPMFeedback
    hutch_shutter: InterlockedHutchShutter
    beamstop: Beamstop
    det_stage: YZStage
    backlight: DualBacklight
    dcm: DCM
